#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__    = 'devsaurus <devsaurus@users.noreply.github.com>'
__copyright__ = 'Copyright 2015'
__license__   = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'

import re
import rrdtool


def plugin(srv, item):

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    # If the incoming payload has been transformed, use that,
    # else the original payload
    text = item.message

    try:
        # addrs is a list[] associated with a particular target.
        # it can contain an arbitrary amount of entries that are just
        # passed along to rrdtool
        # mofified by otfdr @ github to accept abitray arguments with
        # the payload and to not always add the 'N' in front
        # 2017-06-05 - fix/enhancement for https://github.com/mqtt-tools/mqttwarn/issues/248
        if re.match( "^\d+$", text ):
                rrdtool.update(item.addrs, "N:" + text)
        else:
                rrdtool.update(item.addrs + text.split())
    except Exception as e:
        srv.logging.warning("Cannot call rrdtool")
        return False

    return True



# -*- coding: utf-8 -*-

__author__    = 'Tobias Brunner <tobias()tobru.ch>'
__copyright__ = 'Copyright 2016 Tobias Brunner'
__license__   = """Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)"""

import subprocess

def plugin(srv, item):

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    config = item.config
    if type(config) == dict and 'text_replace' in config:
        replace = config['text_replace']
    else:
        replace = '[TEXT]'

    text = item.message
    cmd = [i.replace(replace, text) for i in item.addrs]

    try:
        subprocess.check_call(cmd, stdin=None, stderr=subprocess.STDOUT, shell=False, universal_newlines=True, cwd='/tmp')
    except Exception as e:
        srv.logging.warning("Cannot execute %s because %s" % (cmd, e))
        return False

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Based on the mysql.py service by Jan-Piet Mens

__author__    = 'Jan-Piet Mens <jpmens()gmail.com>, Martyn Whitwell <martyn.whitwell()gmail.com>'
__copyright__ = 'Copyright 2016 Jan-Piet Mens,  Martyn Whitwell'
__license__   = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'


# example configuration in mqttwarn.ini:
# [config:postgres]
# host    = 'localhost'
# user    = 'username'
# pass    = 'password'
# dbname  = 'databasename'
# targets = {
#    'target1': ['table1', 'fallbackcol1', 'schema']
#  }


import psycopg2


def add_row(cursor, tablename, rowdict, schema):
    # XXX tablename not sanitized
    # XXX test for allowed keys is case-sensitive

    unknown_keys = None

    # filter out keys that are not column names
    cursor.execute(
        "SELECT column_name \
        FROM INFORMATION_SCHEMA.COLUMNS \
        where table_schema = '%s' AND table_name = '%s'" % (
        schema, tablename))
    allowed_keys = set(row[0] for row in cursor.fetchall())
    keys = allowed_keys.intersection(rowdict)

    if len(rowdict) > len(keys):
        unknown_keys = set(rowdict) - allowed_keys

    columns = ", ".join(keys)
    values_template = ", ".join(["%s"] * len(keys))

    sql = "insert into %s.%s (%s) values (%s)" % (
        schema, tablename, columns, values_template)
    values = tuple(rowdict[key] for key in keys)
    cursor.execute(sql, values)

    return unknown_keys


def plugin(srv, item):

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    host    = item.config.get('host', 'localhost')
    port    = item.config.get('port', 5432)
    user    = item.config.get('user')
    passwd  = item.config.get('pass')
    dbname  = item.config.get('dbname')

    try:
        table_name = item.addrs[0].format(**item.data)
        fallback_col = item.addrs[1].format(**item.data)
        try:
            schema = item.addrs[2].format(**item.data)
        except:
            schema = 'public'
    except:
        srv.logging.warn("postgres target incorrectly configured")
        return False

    try:
        conn = psycopg2.connect(host=host,
                    port=port,
                    user=user,
                    password=passwd,
                    database=dbname)
        cursor = conn.cursor()
    except Exception as e:
        srv.logging.warn("Cannot connect to postgres: %s" % e)
        return False

    text = item.message

    # Create new dict for column data. First add fallback column
    # with full payload. Then attempt to use formatted JSON values
    col_data = {
        fallback_col : text
       }

    if item.data is not None:
        for key in list(item.data.keys()):
            try:
                col_data[key] = item.data[key].format(**item.data)
            except Exception as e:
                col_data[key] = item.data[key]

    try:
        unknown_keys = add_row(cursor, table_name, col_data, schema)
        if unknown_keys is not None:
            srv.logging.debug("Skipping unused keys %s" % ",".join(unknown_keys))
        conn.commit()
    except Exception as e:
        srv.logging.warn("Cannot add postgres row: %s" % e)
        cursor.close()
        conn.close()
        return False

    cursor.close()
    conn.close()

    return True



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "mqttwarn"
copyright = "2014-2023, Jan-Piet Mens, Ben Jones, Andreas Motl"
author = "Jan-Piet Mens, Ben Jones, Andreas Motl"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx_copybutton",
    "sphinx_togglebutton",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.ifconfig",
    "sphinxext.opengraph",
]


templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]
html_logo = "mqttwarn-logo.png"
html_show_sourcelink = True


# == Extension configuration ==========================================

todo_include_todos = True
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
    "myst": ("https://myst-parser.readthedocs.io/en/latest", None),
}
linkcheck_ignore = [r"https://www.researchgate.net/publication/.*"]
sphinx_tabs_valid_builders = ["linkcheck"]


# -- Options for MyST -------------------------------------------------
myst_heading_anchors = 3
myst_enable_extensions = [
    "attrs_block",
    "attrs_inline",
    "colon_fence",
    "deflist",
    "fieldlist",
    "linkify",
    "strikethrough",
    "tasklist",
]

# -- Options for sphinx-copybutton ------------------------------------
copybutton_remove_prompts = True
copybutton_line_continuation_character = "\\"
copybutton_prompt_text = r">>> |\.\.\. |\$ |sh\$ |PS> |cr> |mysql> |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- Options for sphinxext-opengraph ----------------------------------
ogp_site_url = "https://mqttwarn.readthedocs.io/"
ogp_image = "https://mqttwarn.readthedocs.io/en/latest/_static/mqttwarn-logo.png"
ogp_description_length = 300
ogp_enable_meta_description = True



# -*- coding: utf-8 -*-
# (c) 2014-2023 The mqttwarn developers
from __future__ import print_function

import codecs
import json
import logging
import os
import signal
import sys
import typing as t

from docopt import docopt

from mqttwarn import __version__
from mqttwarn.configuration import Config, load_configuration
from mqttwarn.core import bootstrap, cleanup, run_plugin, subscribe_forever
from mqttwarn.util import get_resource_content

logger = logging.getLogger(__name__)

APP_NAME = "mqttwarn"


def run():
    """
    Usage:
      {program} [make-config]
      {program} [make-udf]
      {program} [--config=] [--config-file=] [--plugin=] [--options=] [--data=]
      {program} --version
      {program} (-h | --help)

    No options:
      mqttwarn will start as a service.

    Interactive options:
      [--config=]               Use configuration settings from JSON string
      [--config-file=]          Use configuration settings from JSON file
      [--plugin=]               The plugin name to load. This can either be a
                                full qualified Python package/module name or a
                                path to a Python file.
      [--options=]              Configuration options to propagate to the plugin entrypoint.
      [--data=]                 Data to propagate to the plugin entrypoint.

    Bootstrapping options:
      make-config               Dump configuration file blueprint to STDOUT,
                                suitable for redirecting into a configuration file.
      make-udf                  Dump blueprint for user-defined functions file to STDOUT,
                                suitable for redirecting into a `udf.py` file.

    Miscellaneous options:
      --version                 Show version information
      -h --help                 Show this screen

    """

    # Use generic commandline options schema and amend with current program name
    commandline_schema = run.__doc__.format(program=APP_NAME)

    # Read commandline options
    options = docopt(commandline_schema, version=APP_NAME + " " + __version__)

    # TODO: Review this. Why do we need it?
    utf8_writer = codecs.getwriter("utf-8")
    sys.stdout = utf8_writer(sys.stdout.buffer)

    if options["make-config"]:
        payload = get_resource_content("mqttwarn.examples", "basic/mqttwarn.ini")
        print(payload)

    elif options["make-udf"]:
        payload = get_resource_content("mqttwarn.examples", "basic/udf.py")
        print(payload)

    elif options["--plugin"] and options["--options"]:

        # Decode arguments
        arg_plugin = options["--plugin"]
        arg_options = options["--options"] and json.loads(options["--options"]) or {}
        arg_data = options["--data"] and json.loads(options["--data"]) or {}
        arg_config = None
        if "--config" in options and options["--config"] is not None:
            arg_config = json.loads(options["--config"])

        # Launch service plugin in standalone mode
        launch_plugin_standalone(
            arg_plugin, arg_options, arg_data, configfile=options.get("--config-file"), config_more=arg_config
        )

    # Run mqttwarn in service mode when no command line arguments are given
    else:
        run_mqttwarn(configfile=options["--config-file"])


def launch_plugin_standalone(
    plugin: str,
    options: t.Dict,
    data: t.Dict,
    configfile: t.Optional[str] = None,
    config_more: t.Optional[t.Dict] = None,
):

    # Optionally load configuration file
    does_not_exist = False
    scriptname = os.path.splitext(os.path.basename(sys.argv[0]))[0]
    try:
        config = load_configuration(configfile=configfile, name=scriptname)
    except FileNotFoundError:
        does_not_exist = True
        config = Config()
        section = "config:{}".format(plugin)
        config.add_section(section)

    # Optionally add additional config settings from command line.
    if config_more is not None:
        section = "config:{}".format(plugin)
        if not config.has_section(section):
            config.add_section(section)
        for key, value in config_more.items():
            config.set(section, key, value)

    # Setup logging
    setup_logging(config)
    if does_not_exist:
        logger.info('Configuration file "{}" does not exist, using default settings'.format(configfile))

    logger.info('Running service plugin "{}" with options "{}"'.format(plugin, options))

    # Launch service plugin
    run_plugin(config=config, name=plugin, options=options, data=data)


def run_mqttwarn(configfile: t.Optional[str] = None):

    # Script name (without extension) used as last resort fallback for config/logfile names
    scriptname = os.path.splitext(os.path.basename(sys.argv[0]))[0]

    # Load configuration file
    config = load_configuration(configfile=configfile, name=scriptname)

    # Setup logging
    setup_logging(config)
    logger.info("Starting {}".format(scriptname))
    logger.info("Log level is %s" % logging.getLevelName(logger.getEffectiveLevel()))

    # Handle signals
    signal.signal(signal.SIGTERM, cleanup)
    signal.signal(signal.SIGINT, cleanup)

    # Bootstrap mqttwarn.core
    bootstrap(config=config, scriptname=scriptname)

    # Connect to broker and start listening
    subscribe_forever()


def setup_logging(config: Config):
    LOGLEVEL = config.loglevelnumber
    LOGFILE = config.logfile
    LOGFORMAT = config.logformat

    # Send log messages to sys.stderr by configuring "logfile = stream://sys.stderr"
    if not LOGFILE:
        pass

    elif LOGFILE.startswith("stream://"):
        LOGFILE = LOGFILE.replace("stream://", "")
        logging.basicConfig(stream=eval(LOGFILE), level=LOGLEVEL, format=LOGFORMAT)

    # Send log messages to file by configuring "logfile = 'mqttwarn.log'"
    else:
        logging.basicConfig(filename=LOGFILE, level=LOGLEVEL, format=LOGFORMAT)



from mqttwarn.model import Service, ProcessorItem


def plugin(srv: Service, item: ProcessorItem) -> bool:
    """
    An mqttwarn plugin with little overhead, suitable for unit- and integration-testing.
    """
    if hasattr(item, "message") and item.message and "fail" in item.message:
        srv.logging.error("Failed sending message using noop")
        return False
    else:
        srv.logging.info("Successfully sent message using noop")
        return True



# -*- coding: utf-8 -*-
# (c) 2021-2023 The mqttwarn developers
import dataclasses
import platform
import sys
import typing as t
from dataclasses import dataclass, field
from functools import total_ordering
from logging import Logger
from typing import Dict, Optional, Union

from mqttwarn import __version__

# Type definitions.

# The venerable transformation data dictionary.
TdataType = t.Dict[str, t.Union[t.AnyStr, int]]

# Covering old- and new-style configuration layouts. `addrs` has
# originally been a list of strings, has been expanded to be a
# list of dictionaries (Apprise), to be a dictionary (Pushsafer),
# and finally to be a scalar string only (ntfy).
TopicTargetType = t.Union[t.List, t.Dict[str, t.Any], str, None]


class Struct:
    """
    Data container for feeding information into service plugins - V1.
    """

    # Convert Python dict to object?
    # http://stackoverflow.com/questions/1305532/

    def __init__(self, **entries):
        self.__dict__.update(entries)

    def __repr__(self):
        return "<%s>" % str("\n ".join("%s: %s" % (k, repr(v)) for (k, v) in list(self.__dict__.items())))

    def get(self, key, default=None):
        if key in self.__dict__ and self.__dict__[key] is not None:
            return self.__dict__[key]
        else:
            return default

    def enum(self):
        item = {}
        for (k, v) in list(self.__dict__.items()):
            item[k] = v
        return item


@dataclass
class ProcessorItem:
    """
    Data container for feeding information into service plugins - V2.
    """

    service: Optional[str] = None
    target: Optional[str] = None
    config: Dict = field(default_factory=dict)
    section: Optional[str] = None
    # TODO: `addrs` can also be a string or dictionary now.
    addrs: TopicTargetType = field(default_factory=list)  # type: ignore[assignment]
    priority: Optional[int] = None
    topic: Optional[str] = None
    title: Optional[str] = None
    message: Optional[Union[str, bytes]] = None
    data: Optional[Dict] = None

    def asdict(self):
        return dataclasses.asdict(self)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def to_job(self) -> "Job":
        return Job(
            prio=self.priority,
            service=self.service,
            section=self.section,
            topic=self.topic,
            payload=self.message,
            data=self.data,
            target=self.target,
        )


@dataclasses.dataclass
class StatusInformation:
    """
    Different bits of information published to `mqttwarn/$SYS` when the `status_publish` feature is enabled.
    """

    mqttwarn_version: str = __version__
    os_platform: str = sys.platform
    python_version: str = platform.python_version()


class Service:
    """
    Class with helper functions which is passed to each plugin and its global instantiation.
    """

    def __init__(self, mqttc, logger, mwcore, program):

        # Reference to MQTT client object.
        self.mqttc = mqttc

        # Reference to all mqttwarn globals, for using its machinery from plugins.
        self.mwcore = mwcore

        # Reference to logging object.
        self.logging: Logger = logger  # type: ignore[annotation-unchecked]

        # Name of self ("mqttwarn", mostly).
        self.SCRIPTNAME = program


@total_ordering
class Job:
    def __init__(self, prio, service, section, topic, payload, data, target):
        self.prio = prio
        self.service = service
        self.section = section
        self.topic = topic
        self.payload = payload  # raw payload
        self.data = data  # decoded payload
        self.target = target

    # The `__cmp__()` special method is no longer honored in Python 3.
    # https://portingguide.readthedocs.io/en/latest/comparisons.html#rich-comparisons

    def __eq__(self, other):
        return self.prio == other.prio

    def __ne__(self, other):
        return not (self.prio == other.prio)

    def __lt__(self, other):
        return self.prio < other.prio



import json


def togglestate(topic, payload, section, srv):
    filename = "warntoggle.json"
    default_topicblock = False

    try:
        with open(filename) as infile:
            toggles = json.load(infile)
            infile.close()

        if topic in toggles:
            # file found, topic found
            topicblock = toggles[topic]
            srv.logging.debug('togglestate() was called from the {} section and found {} in {}'.format(
                section, topic, filename))
        else:
            # file found, adding new topic
            toggles[topic] = default_topicblock
            with open(filename, 'w') as outfile:
                json.dump(toggles, outfile)
                outfile.close()

            topicblock = default_topicblock
            srv.logging.debug('togglestate() was called from the {} section, did not find {} in {}'.format(
                section, topic, filename))
            srv.logging.debug('togglestate() added {} to {} with blocking set to {}'.format(
                topic, filename, topicblock))

    except Exception as e:
        # file not found or other error
        topicblock = default_topicblock
        srv.logging.debug('togglestate() encountered an error: {}'.format(e))

    srv.logging.debug('togglestate() will return {}'.format(topicblock))
    return topicblock



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__    = 'Joerg Gollnick <github+mqttwarn()wurzelbenutzer.de>'
__copyright__ = 'Copyright 2016 Tobias Brunner / 2021 Joerg Gollnick'
__license__   = """Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)"""

import subprocess
import json
from pipes import quote
from six import string_types

def plugin(srv, item):

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    # same as for ssh
    json_message=json.loads(item.message)

    args = None
    if json_message is not None:
        args = json_message["args"]

        if type(args) is list and len(args) == 1:
            args=args[0]

        if type(args) is list:
            args=tuple([ quote(v) for v  in args ]) #escape the shell args
        elif type(args) is str or type(args) is unicode:
            args=(quote(args),)

    # parse topic
    topic=list(map( lambda x: quote(x), item.topic.split('/') ))

    outgoing_topic = None
    # replace palceholders args[0], args[1] ..., full_topic, topic[0],
    if item.addrs[0] is not None:
        outgoing_topic = item.addrs[0].format(args=args,full_topic=quote(item.topic),topic=topic)
    qos            = item.addrs[1]
    retain         = item.addrs[2]
    addrs          = item.addrs[3:]

    cmd = None
    if addrs is not None:
        cmd = [i.format(args=args, full_topic=quote(item.topic),topic=topic) for i in addrs]

    srv.logging.debug("*** MODULE=%s: service=%s, command=%s outgoing_topic=%s", __file__, item.service, str( cmd ),outgoing_topic)

    try:
        res = subprocess.check_output(cmd, stdin=None, stderr=subprocess.STDOUT, shell=False, universal_newlines=True, cwd='/tmp')
    except Exception as e:
        srv.logging.warning("Cannot execute %s because %s" % (cmd, e))
        return False

    if outgoing_topic is not None:
        outgoing_payload = res.rstrip('\n')
        if isinstance(outgoing_payload, string_types):
            outgoing_payload = bytearray(outgoing_payload, encoding='utf-8')
        try:
            srv.mqttc.publish(outgoing_topic, outgoing_payload, qos=qos, retain=retain)
        except Exception as e:
            srv.logging.warning("Cannot PUBlish response %s: %s" % (outgoing_topic, e))
            return False

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__    = 'Jan-Piet Mens <jpmens()gmail.com>'
__copyright__ = 'Copyright 2014 Jan-Piet Mens'
__license__   = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'

import dataclasses
from collections import OrderedDict

from future import standard_library
standard_library.install_aliases()
import os
import json
import urllib.request, urllib.parse, urllib.error
import urllib.request, urllib.error, urllib.parse
import urllib.parse

import typing as t

from mqttwarn.model import ProcessorItem

PUSHSAFER_API = "https://www.pushsafer.com/"


class PushsaferError(Exception): pass
class PushsaferConfigurationError(Exception): pass


def pushsafer(**kwargs):
    """
    Submit notification to Pushsafer.
    """
    assert 'm' in kwargs

    if not kwargs['k']:
        kwargs['k'] = os.environ['PUSHSAFER_TOKEN']

    # Don't submit empty parameters to Pushsafer.
    filter_empty_parameters(kwargs)

    url = urllib.parse.urljoin(PUSHSAFER_API, "api")
    data = urllib.parse.urlencode(kwargs).encode('utf-8')
    req = urllib.request.Request(url, data)
    response = urllib.request.urlopen(req, timeout=3)
    output = response.read()
    data = json.loads(output)

    if data['status'] != 1:
        raise PushsaferError(output)


def filter_empty_parameters(params: t.Dict[str, str]):
    """
    Filter empty parameters.
    """
    for key in list(params.keys()):
        value = params[key]
        if value == "":
            del params[key]


def plugin(srv, item):

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    # Compute Pushsafer parameters from `ProcessorItem`.
    pe = PushsaferParameterEncoder(item=item)
    try:
        pe.encode()
    except PushsaferConfigurationError as ex:
        srv.logging.error(f"{ex}. target={item.target}")
        return False

    # Submit notification.
    params = pe.params
    try:
        srv.logging.debug("Sending pushsafer notification to %s [%s]..." % (item.target, params))
        pushsafer(m=item.message, k=pe.private_key, **params)
        srv.logging.debug("Successfully sent pushsafer notification")
    except Exception as e:
        srv.logging.warning("Error sending pushsafer notification to %s [%s]: %s" % (item.target, params, e))
        return False

    return True


class PushsaferParameterEncoder:
    """
    Encode Pushsafer parameters from mqttwarn configuration layout.
    """
    def __init__(self, item: ProcessorItem):
        self.item = item
        self.private_key: t.Optional[str] = None
        self.params: t.Dict[str, str] = {}

    def encode(self):
        addrs = self.item.addrs
        if isinstance(addrs, t.List):
            self.encode_v1()
        elif isinstance(addrs, t.Dict):
            self.encode_v2()
        else:
            raise ValueError(f"Pushsafer configuration layout empty or invalid. type={type(addrs).__name__}")

    def encode_v1(self):
        """
        Based on the Pushsafer API (https://www.pushsafer.com/en/pushapi).
        `addrs` is a list with two or three elements.

        0 is the private or alias key
        1 (if present) is the Pushsafer device or device group id where the message is to be sent
        2 (if present) is the Pushsafer icon to display in the message
        3 (if present) is the Pushsafer sound to play for the message
        4 (if present) is the Pushsafer vibration for the message
        5 (if present) is the Pushsafer URL or URL Scheme
        6 (if present) is the Pushsafer Title of URL
        7 (if present) is the Pushsafer Time in minutes, after which message automatically gets purged
        8 (if present) is the Pushsafer priority, integer -2, -1, 0, 1, 2
        9 (if present) is the Pushsafer retry after which time (in seconds 60-10800) a message should resend
        10 (if present) is the Pushsafer expire after which time (in seconds 60-10800) the retry should stopped
        11 (if present) is the Pushsafer answer, 1 = Answer, 0 = no possibility to answer
        12 (if present) is the Pushsafer answer options seperated by a pipe character e.g. yes|no|maybe
        13 (if present) is the Pushsafer force answer, 1 = yes, 0 = no
        14 (if present) is the Pushsafer message will be repeated after specified time delay when not confirmed
        """

        addrs = self.item.addrs
        title = self.item.title

        # Decode Private or Alias Key.
        try:
            self.private_key = addrs[0]
        except IndexError:
            raise PushsaferConfigurationError(f"Pushsafer private or alias key not configured")

        params = OrderedDict()

        if len(addrs) > 1:
            params['d'] = addrs[1]

        if len(addrs) > 2:
            params['i'] = addrs[2]

        if len(addrs) > 3:
            params['s'] = addrs[3]

        if len(addrs) > 4:
            params['v'] = addrs[4]

        if len(addrs) > 5:
            params['u'] = addrs[5]

        if len(addrs) > 6:
            params['ut'] = addrs[6]

        if len(addrs) > 7:
            params['l'] = addrs[7]

        if len(addrs) > 8:
            params['pr'] = addrs[8]

        if len(addrs) > 9:
            params['re'] = addrs[9]

        if len(addrs) > 10:
            params['ex'] = addrs[10]

        if len(addrs) > 11:
            params['a'] = addrs[11]

        if len(addrs) > 12:
            params['ao'] = addrs[12]

        if len(addrs) > 13:
            params['af'] = addrs[13]

        if len(addrs) > 14:
            params['cr'] = addrs[14]

        if title is not None:
            params['t'] = title

        self.params = params

    def encode_v2(self):
        """
        New-style configuration layout with named parameters for Pushsafer.
        """

        addrs = self.item.addrs
        title = self.item.title

        # Decode Private or Alias Key.
        try:
            self.private_key = addrs["private_key"]
        except KeyError:
            raise PushsaferConfigurationError(f"Pushsafer private or alias key not configured")

        # Decode and serialize all other parameters.
        pp = PushsaferParameters(**addrs)
        params = pp.translated()

        # Propagate `title` separately.
        if title is not None:
            params['t'] = title

        self.params = params


@dataclasses.dataclass
class PushsaferParameters:
    """
    Manage available Pushsafer parameters, and map them to their short representations,
    suitable for sending over the wire.

    - https://www.pushsafer.com/en/pushapi
    - https://www.pushsafer.com/en/pushapi_ext
    """
    private_key: t.Optional[str] = None
    device: t.Optional[str] = None
    icon: t.Optional[int] = None
    sound: t.Optional[int] = None
    vibration: t.Optional[int] = None
    url: t.Optional[str] = None
    url_title: t.Optional[str] = None
    time_to_live: t.Optional[int] = None
    priority: t.Optional[int] = None
    retry: t.Optional[int] = None
    expire: t.Optional[int] = None
    answer: t.Optional[int] = None
    answer_options: t.Optional[str] = None
    answer_force: t.Optional[int] = None
    confirm_repeat: t.Optional[int] = None

    PARAMETER_MAP = {
        "device": "d",
        "icon": "i",
        "sound": "s",
        "vibration": "v",
        "url": "u",
        "url_title": "ut",
        "time_to_live": "l",
        "priority": "pr",
        "retry": "re",
        "expire": "ex",
        "answer": "a",
        "answer_options": "ao",
        "answer_force": "af",
        "confirm_repeat": "cr",
    }

    def to_dict(self) -> t.Dict[str, t.Union[str, int]]:
        return dataclasses.asdict(self)

    def translated(self) -> t.Dict[str, t.Union[str, int]]:
        """
        Translate parameters to their wire representations.
        """
        result = OrderedDict()
        for attribute, parameter_name in self.PARAMETER_MAP.items():
            value = getattr(self, attribute)
            if value is not None:
                result[parameter_name] = value
        return result



import logging

import pytest

from mqttwarn.model import Service


@pytest.fixture
def mqttwarn_service():
    """
    A service instance for propagating to the plugin.
    """
    logger = logging.getLogger(__name__)
    # FIXME: Should propagate `mqttwarn.core.globals()` to `mwcore`.
    return Service(mqttc=None, logger=logger, mwcore={}, program="mqttwarn-testdrive")



def decode_for_zabbix(topic, data, srv=None):
    status_key = None

    # the first part (part[0]) is always tele
    # the second part (part[1]) is the device, the value comes from
    # the third part (part[2]) is the name of the metric (e.g. temperature/humidity/voltage...)
    parts = topic.split('/')
    client = parts[1]
    key = parts[2]

    return dict(client=client, key=key, status_key=status_key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__author__    = 'Bram Hendrickx'
__copyright__ = 'Copyright 2016 Bram Hendrickx'
__license__   = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'

Original script by Bram Hendrickx. https://github.com/mqtt-tools/mqttwarn/blob/main/mqttwarn/services/ifttt.py
Modified to work with hangoutsbot api plugin https://github.com/hangoutsbot/hangoutsbot/wiki/API-Plugin
"""

__author__    = 'Michael Brougham'
__copyright__ = 'Copyright 2018 Michael Brougham'
__license__   = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'


import json
import requests


def plugin(srv, item):
    """ expects (url,port,apikey, convid) in addrs """

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    payload = {
        'key': item.addrs[2],
        'sendto': item.addrs[3],
        'content': item.message
    }
    try:
        srv.logging.debug("Sending to hangoutsbot")
        url = "https://" + item.addrs[0] + ":" + item.addrs[1]
        headers = {'content-type': 'application/json'}
        requests.post(url, data = json.dumps(payload), headers = headers, verify=False)
        srv.logging.debug("Successfully sent to hangoutsbot")
    except Exception as e:
        srv.logging.warning("Failed to send message to hangoutsbot" % e)
        return False

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__    = 'Fabian Affolter <fabian()affolter-engineering.ch>'
__copyright__ = 'Copyright 2014 Fabian Affolter'
__license__   = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'

from gi.repository import Notify


def plugin(srv, item):
    """Send a message to the user's desktop notification system."""

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__,
            item.service, item.target)

    title = item.addrs[0]
    text = item.message

    try:
        srv.logging.debug("Sending notification to the user's desktop")
        Notify.init('mqttwarn')
        n = Notify.Notification.new(
            title,
            text,
            '/usr/share/icons/gnome/32x32/places/network-server.png')
        n.show()
        srv.logging.debug("Successfully sent notification")
    except Exception as e:
        srv.logging.warning("Cannot invoke notification to linux: %s" % e)
        return False

    return True



# -*- coding: utf-8 -*-
# (c) 2014-2023 The mqttwarn developers
import logging
import os
import socket
import sys
import threading
import time
import typing as t
from builtins import chr, str
from datetime import datetime
from queue import Queue

import paho.mqtt.client as paho
from paho.mqtt.client import Client as MqttClient
from paho.mqtt.client import MQTTMessage
from pkg_resources import resource_filename

import mqttwarn.configuration
from mqttwarn.context import FunctionInvoker, RuntimeContext
from mqttwarn.cron import PeriodicThread
from mqttwarn.model import Job, Service, StatusInformation, Struct, TdataType
from mqttwarn.util import (
    Formatter,
    asbool,
    load_function,
    load_module_by_name,
    load_module_from_file,
    parse_cron_options,
    sanitize_function_name,
    timeout,
    truncate,
)

try:
    import json
except ImportError:  # pragma: nocover
    import simplejson as json  # type: ignore


HAVE_JINJA = True
try:
    from jinja2 import Environment, FileSystemLoader

    jenv = Environment(loader=FileSystemLoader("templates/", encoding="utf-8"), trim_blocks=True)
    jenv.filters["jsonify"] = json.dumps
except ImportError:
    HAVE_JINJA = False


logger = logging.getLogger(__name__)

# Name of calling program
SCRIPTNAME = "mqttwarn"

# Global runtime context object
context: RuntimeContext
context = None  # type: ignore[assignment]

# Global configuration object
cf: mqttwarn.configuration.Config
cf = None  # type: ignore[assignment]

# Global handle to MQTT client
mqttc: paho.Client
mqttc = None  # type: ignore[assignment]

# Initialize processor queue
q_in: Queue = Queue(maxsize=0)
exit_flag = False

# Instances of PeriodicThread objects
ptlist: t.Dict[str, PeriodicThread] = {}

# Instances of loaded service plugins
service_plugins: t.Dict[str, t.Dict[str, t.Any]] = dict()


def make_service(mqttc: paho.Client = None, name: t.Optional[str] = None) -> Service:
    """
    Service object factory.
    Prepare service object for plugin.
    Inject appropriate MQTT client and logger objects.

    :param mqttc: Instance of PAHO MQTT client object.
    :param name:  Name used for obtaining a logger instance.
    :return:      Service object ready for being passed to plugin instance.
    """
    name = name or "unknown"
    logger = logging.getLogger(name)
    service = Service(mqttc=mqttc, logger=logger, mwcore=globals(), program=SCRIPTNAME)
    return service


def render_template(filename: str, data: TdataType) -> t.Optional[str]:
    text = None
    if HAVE_JINJA is True:
        template = jenv.get_template(filename)
        text = template.render(data)

    return text


# MQTT broker callbacks
def on_connect(mosq: MqttClient, userdata: t.Dict[str, str], flags: t.Dict[str, str], result_code: int):
    """
    Handle connections (or failures) to the broker.
    This is called after the client has received a CONNACK message
    from the broker in response to calling connect().

    The result_code is one of;
    0: Success
    1: Refused - unacceptable protocol version
    2: Refused - identifier rejected
    3: Refused - server unavailable
    4: Refused - bad user name or password (MQTT v3.1 broker only)
    5: Refused - not authorised (MQTT v3.1 broker only)
    """
    if result_code == 0:
        logger.debug("Connected to MQTT broker, subscribing to topics")
        if not cf.cleansession:
            logger.debug(
                "Cleansession==False; previous subscriptions for clientid %s remain active on broker" % cf.clientid
            )

        subscribed = []
        for section in context.get_sections():
            topic = context.get_topic(section)
            qos = context.get_qos(section)

            if topic in subscribed:
                continue

            logger.debug("Subscribing to %s (qos=%d)" % (topic, qos))
            mqttc.subscribe(topic, qos)
            subscribed.append(topic)

        if cf.lwt is not None:
            mqttc.publish(cf.lwt, cf.lwt_alive, qos=0, retain=True)

    elif result_code == 1:
        logger.error("Connection refused - unacceptable protocol version")
    elif result_code == 2:
        logger.error("Connection refused - identifier rejected")
    elif result_code == 3:
        logger.error("Connection refused - server unavailable")
    elif result_code == 4:
        logger.error("Connection refused - bad user name or password")
    elif result_code == 5:
        logger.error("Connection refused - not authorised")
    else:
        logger.error("Connection failed - result code %d" % (result_code))


def on_disconnect(mosq: MqttClient, userdata: t.Dict[str, str], result_code: int):
    """
    Handle disconnections from the broker
    """
    if result_code == 0:
        logger.info("Clean disconnection from broker")
    else:
        send_failover("brokerdisconnected", b"Broker connection lost. Will attempt to reconnect in 5s")
        # TODO: Review this.
        time.sleep(5)


def on_message(mosq: MqttClient, userdata: t.Dict[str, str], msg: MQTTMessage):
    """
    Dispatch message received from the MQTT broker to mqttwarn's handler machinery.
    """
    try:
        return on_message_handler(mosq, userdata, msg)
    except:
        logger.exception("Receiving and decoding MQTT message failed")


def on_message_handler(mosq: MqttClient, userdata: t.Dict[str, str], msg: MQTTMessage):
    """
    Message received from the broker
    """

    topic = msg.topic
    payload = msg.payload
    logger.debug(f"Message received on {topic}: {truncate(payload)}")

    if msg.retain == 1:
        if cf.skipretained:
            logger.debug("Skipping retained message on %s" % topic)
            return

    # Try to find matching settings for this topic
    for section in context.get_sections():
        # Get the topic for this section (usually the section name but optionally overridden)
        match_topic = context.get_topic(section)
        if paho.topic_matches_sub(match_topic, topic):
            logger.debug("Section [%s] matches message on %s, processing it" % (section, topic))
            # Check for any message filters
            if context.is_filtered(section, topic, payload):
                logger.log(
                    cf.filteredmessagesloglevelnumber,
                    "Filter in section [%s] has skipped message on %s" % (section, topic),
                )
                continue
            # Send the message to any targets specified
            send_to_targets(section, topic, payload)


# End of MQTT broker callbacks


def send_failover(reason: str, message: t.AnyStr):
    # Make sure we dump this event to the log
    logger.warning(message)
    # Attempt to send the message to our failover targets
    send_to_targets("failover", reason, message)


def send_to_targets(section: str, topic: str, payload: t.AnyStr):
    if cf.has_section(section) is False:
        logger.warning(
            "Section [%s] does not exist in your INI file, skipping message on topic '%s'" % (section, topic)
        )
        return

    # decode raw payload into transformation data
    data = decode_payload(section, topic, payload)

    # Probe if it's a function name.
    function_name = None
    try:
        function_name = sanitize_function_name(context.get_config(section, "targets"))
    except:
        pass
    dispatcher_dict = cf.getdict(section, "targets")

    # `targets` is a function symbol.
    if function_name is not None:
        targetlist = context.get_topic_targets(section, topic, data)

        # Make sure the function returned a target _list_ of elements.
        if not isinstance(targetlist, list):
            targetlist_type = type(targetlist)
            logger.error(
                'Topic target definition by function "{function_name}" '
                'in section "{section}" is empty or incorrect. Should be a list. '
                "targetlist={targetlist}, type={targetlist_type}".format(**locals())
            )
            return

    # `targets` is a dictionary.
    elif isinstance(dispatcher_dict, dict) and dispatcher_dict:

        def get_key(item):
            # precede a key with the number of topic levels and then use reverse alphabetic sort order
            # '+' is after '#' in ascii table
            # caveat: for instance space is allowed in topic name but will be less specific than '+', '#'
            # so replace '#' with first ascii character and '+' with second ascii character
            # http://public.dhe.ibm.com/software/dw/webservices/ws-mqtt/mqtt-v3r1.html#appendix-a

            # item[0] represents topic. replace wildcard characters to ensure the right order
            modified_topic = item[0].replace("#", chr(0x01)).replace("+", chr(0x02))
            levels = len(item[0].split("/"))
            # concatenate levels with leading zeros and modified topic and return as a key
            return "{:03d}{}".format(levels, modified_topic)

        # produce a sorted list of topic/targets with longest and more specific first
        sorted_dispatcher = sorted(list(dispatcher_dict.items()), key=get_key, reverse=True)
        for match_topic, targets in sorted_dispatcher:
            if paho.topic_matches_sub(match_topic, topic):
                # hocus pocus, let targets become a list
                targetlist = targets if isinstance(targets, list) else [targets]
                logger.debug("Most specific match %s dispatched to %s" % (match_topic, targets))
                # first most specific topic matches then stops processing
                break
        else:
            # Not found then no action. This could be configured intentionally.
            logger.debug("Dispatcher definition does not contain matching topic/target pair in section [%s]" % section)
            return

    else:
        targetlist = cf.getlist(section, "targets")

        # Make sure targets are actually a _list_ of elements.
        # TODO: Not tested yet. How can this code be reached?
        if targetlist is None:
            logger.error(
                f'Topic target definition in section "{section}" is empty or incorrect. Should be a list. '
                "targetlist={targetlist}".format(**locals())
            )
            return

    # Interpolate transformation data values into topic targets.
    # Be graceful if interpolation fails, but log a meaningful message.
    targetlist_resolved = []
    for target in targetlist:
        try:
            target = target.format(**data)
            targetlist_resolved.append(target)
        except Exception as ex:
            error = repr(ex)
            logger.error(
                f"Interpolating transformation data into topic target '{target}' failed. Reason: {error}. "
                f"section={section}, topic={topic}, payload={truncate(payload)}, data=%s",
                data,
            )
    targetlist = targetlist_resolved

    for item in targetlist:
        logger.debug("Message on %s going to %s" % (topic, item))
        # Each target is either "service" or "service:target"
        # If no target specified then notify ALL targets
        service = item
        target = None

        # Check if this is for a specific target
        if item.find(":") != -1:
            try:
                service, target = item.split(":", 2)
            except:
                logger.error(f"Invalid topic target: {item}. Should be 'service:target'.")
                continue

        # Skip targets with invalid services.
        if service not in service_plugins:
            logger.error("Invalid configuration: Topic '%s' points to non-existing service '%s'" % (topic, service))
            continue

        service_config = context.get_service_config(service)
        payload_out: t.Union[str, bytes]
        if asbool(service_config.get("decode_utf8", True)) and isinstance(payload, bytes):
            payload_out = payload.decode("utf-8")
        else:
            payload_out = payload

        sendtos = None
        if target is None:
            sendtos = context.get_service_targets(service)
        else:
            sendtos = [target]

        for sendto in sendtos:
            logger.debug("New `%s:%s' job: %s" % (service, sendto, topic))
            job = Job(1, service, section, topic, payload_out, data, sendto)
            q_in.put(job)


def builtin_transform_data(topic: str, payload: t.AnyStr) -> TdataType:
    """Return a dict with initial transformation data which is made
    available to all plugins"""

    tdata: TdataType = {}
    dt = datetime.now()

    tdata["topic"] = topic
    tdata["payload"] = payload
    tdata["_dtepoch"] = int(time.time())  # 1392628581
    tdata["_dtiso"] = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%fZ")  # 2014-02-17T10:38:43.910691Z
    tdata["_ltiso"] = datetime.now().isoformat()  # local time in iso format
    tdata["_dthhmm"] = dt.strftime("%H:%M")  # 10:16
    tdata["_dthhmmss"] = dt.strftime("%H:%M:%S")  # hhmmss=10:16:21

    return tdata


def xform(function: str, orig_value: t.Any, transform_data: TdataType) -> t.Union[TdataType, str, None]:
    """
    Attempt transformation on orig_value.

    - 1st. function()
    - 2nd. inline {xxxx}
    """

    if orig_value is None:
        return None

    res = orig_value

    if function is not None:
        try:
            function_name = sanitize_function_name(function)
            try:
                res = context.invoker.datamap(function_name, transform_data)
                return res
            except:
                logger.exception(f"Invoking function failed: {function}")
        except:
            pass

        try:
            res = Formatter().format(function, **transform_data)
        except:
            logger.exception(f"Formatting message with function failed: {function}")

    if isinstance(res, str):
        res = res.replace("\\n", "\n")

    return res


def decode_payload(section: str, topic: str, payload: t.AnyStr) -> TdataType:
    """
    Decode message payload through transformation machinery.
    """

    transform_data = builtin_transform_data(topic, payload)

    topic_data = context.get_topic_data(section, transform_data)
    if topic_data is not None and isinstance(topic_data, dict):
        transform_data.update(topic_data)

    # The dict returned is completely merged into transformation data
    # The difference between this and `get_topic_data()' is that this
    # function obtains the topic string as well as the payload and any
    # existing transformation data, and it can do 'things' with all.
    # This is the way it should originally have been, but I can no
    # longer fix the original (legacy).

    all_data = context.get_all_data(section, topic, transform_data)
    if all_data is not None and isinstance(all_data, dict):
        transform_data.update(all_data)

    # Gracefully attempt to decode the payload from JSON. If it's possible, add
    # the JSON keys into item to pass to the plugin, and create the outgoing
    # (i.e. transformed) message.
    try:
        if isinstance(payload, bytes):
            outdata = payload.decode("utf-8")
        else:
            outdata = payload
        outdata = outdata.rstrip("\0")
        payload_data = json.loads(outdata)
        transform_data.update(payload_data)
    except Exception as ex:
        logger.debug(f"Decoding JSON failed: {ex}. payload={truncate(payload)}")

    return transform_data


def processor(worker_id=None):
    """
    Queue runner. Pull a job from the queue, find the module in charge
    of handling the service, and invoke the module's plugin to do so.
    """
    while not exit_flag:
        logger.debug("Job queue has %s items to process" % q_in.qsize())
        job = q_in.get()
        if process_job(job=job, worker_id=worker_id):
            q_in.task_done()
    logger.debug("Worker thread exiting")


def process_job(job, worker_id=None):
    """
    Process a single job item.
    """

    if True:

        service = job.service
        section = job.section
        target = job.target
        topic = job.topic

        logger.debug("Processor #%s is handling: `%s' for %s" % (worker_id, service, target))

        # Sanity checks.
        # If service configuration or targets can not be obtained successfully,
        # log a sensible error message, fail the job, and carry on with the next job.
        try:
            service_config = context.get_service_config(service)
            service_targets = context.get_service_targets(service)

            if target not in service_targets:
                error_message = (
                    "Invalid configuration: Topic '{topic}' points to "
                    "non-existing target '{target}' in service '{service}'".format(**locals())
                )
                raise KeyError(error_message)

        except Exception:
            logger.exception(f"Cannot handle service={service}, target={target}")
            return True

        # Be more graceful with jobs w/o any target address information (2021-10-18 [amo]).
        if target is None:
            addrs = []
        else:
            addrs = service_targets[target]

        item = {
            "service": service,
            "section": section,
            "target": target,
            "config": service_config,
            "addrs": addrs,
            "topic": topic,
            "payload": job.payload,
            "data": None,
            "title": None,
            "image": None,
            "message": None,
            "priority": None,
        }

        transform_data = job.data
        item["data"] = dict(list(transform_data.items()))

        origin_title = "{}: {}".format(SCRIPTNAME, topic)
        item["title"] = xform(context.get_config(section, "title"), origin_title, transform_data)
        item["image"] = xform(context.get_config(section, "image"), "", transform_data)
        item["message"] = xform(context.get_config(section, "format"), job.payload, transform_data)

        try:
            item["priority"] = int(xform(context.get_config(section, "priority"), 0, transform_data))
        except:
            item["priority"] = 0
            logger.exception("Failed to determine the priority, defaulting to zero")

        if HAVE_JINJA is False and context.get_config(section, "template"):
            logger.warning("Templating not possible because Jinja2 is not installed")

        if HAVE_JINJA is True:
            template = context.get_config(section, "template")
            if template is not None:
                try:
                    text = render_template(template, transform_data)
                    if text is not None:
                        item["message"] = text
                except:
                    logger.exception(f"Rendering template failed: {template}")

        if item.get("message") is not None and len(item.get("message")) > 0:
            st = Struct(**item)
            notified = False
            logger.info("Invoking service plugin for `%s'" % service)
            try:
                # Fire the plugin in a separate thread and kill it if it doesn't return in 10s
                module = service_plugins[service]["module"]
                if "." in service:
                    service_logger_name = service
                else:
                    service_logger_name = "mqttwarn.services.{}".format(service)
                srv = make_service(mqttc=mqttc, name=service_logger_name)
                notified = timeout(module.plugin, (srv, st))
            except Exception as ex:
                logger.exception(f"Invoking service failed. Reason: {ex}. service={service}, topic={topic}")

            if not notified:
                logger.warning(f"Notification failed or timed out. service={service}, topic={topic}")
        else:
            logger.info(f"Notification suppressed. Reason: Payload is empty. service={service}, topic={topic}")

        return True


def load_services(services):

    if services is None:
        logger.warning("No services defined")
        return

    for service in services:
        service_plugins[service] = {}

        service_config = cf.config("config:" + service)
        service_plugins[service]["config"] = service_config

        module = cf.g("config:" + service, "module", service)

        # Load external service from file.
        modulefile_candidates = []
        if module.endswith(".py"):
            # Add two candidates: a) Use the file as given and b) treat the file as relative to
            # the directory of the configuration file. That retains backward compatibility.
            modulefile_candidates.append(module)
            modulefile_candidates.append(os.path.join(cf.configuration_path, module))

        # Load external service with module specification.
        elif "." in module:
            logger.debug('Trying to load service "{}" from module "{}"'.format(service, module))
            try:
                service_plugins[service]["module"] = load_module_by_name(module)
                logger.info('Successfully loaded service "{}" from module "{}"'.format(service, module))
                continue
            except:
                logger.exception('Loading service "{}" from module "{}" failed'.format(service, module))

        # Load built-in service module.
        else:
            # Backward-compatibility patch for honoring the renaming of the `http.py` module.
            if module == "http":
                module = "http_urllib"
            logger.debug('Trying to load built-in service "{}" from "{}"'.format(service, module))
            modulefile_candidates = [resource_filename("mqttwarn.services", module + ".py")]

        success = False
        for modulefile in modulefile_candidates:
            if not os.path.isfile(modulefile):
                logger.error('Module "{}" is not a file'.format(modulefile))
                continue
            logger.debug('Trying to load service "{}" from file "{}"'.format(service, modulefile))
            try:
                service_plugins[service]["module"] = load_module_from_file(modulefile)
                logger.info('Successfully loaded service "{}"'.format(service))
                success = True
            except:
                logger.exception(f'Loading service "{service}" from file "{modulefile}" failed')

        if not success:
            logger.critical('Unable to load service "{}"'.format(service))
            # TODO: Review this.
            sys.exit(1)


def connect():
    """
    Load service plugins, connect to the MQTT broker, launch daemon threads, and listen forever.
    """

    # FIXME: Remove global variables
    global mqttc

    try:
        services = cf.getlist("defaults", "launch")
    except:
        logger.error("No services configured, aborting")
        # TODO: Review this.
        sys.exit(2)

    load_services(services)

    # Initialize MQTT broker connection
    mqttc = paho.Client(cf.clientid, clean_session=cf.cleansession, protocol=cf.protocol)

    logger.debug("Attempting connection to MQTT broker %s:%d" % (cf.hostname, int(cf.port)))
    mqttc.on_connect = on_connect
    mqttc.on_message = on_message
    mqttc.on_disconnect = on_disconnect

    # check for authentication
    if cf.username:
        mqttc.username_pw_set(cf.username, cf.password)

    # set the lwt before connecting
    if cf.lwt is not None:
        logger.debug("Setting LWT to %s" % (cf.lwt))
        mqttc.will_set(cf.lwt, payload=cf.lwt_dead, qos=0, retain=True)

    # Delays will be: 3, 6, 12, 24, 30, 30, etc.
    # mqttc.reconnect_delay_set(delay=3, delay_max=30, exponential_backoff=True)

    if cf.tls is True:
        mqttc.tls_set(cf.ca_certs, cf.certfile, cf.keyfile, tls_version=cf.tls_version, ciphers=None)

    if cf.tls_insecure:
        mqttc.tls_insecure_set(True)

    try:
        mqttc.connect(cf.hostname, int(cf.port), 60)

    except Exception:
        logger.exception("Cannot connect to MQTT broker at %s:%d" % (cf.hostname, int(cf.port)))
        # TODO: Review this.
        sys.exit(2)

    # Update our runtime context (used by functions etc) now we have a connected MQTT client
    context.invoker.srv.mqttc = mqttc

    # Publish status information to `mqttwarn/$SYS` topic.
    publish_status_information()

    # Launch worker threads to operate on queue
    start_workers()

    return mqttc


def subscribe_forever():
    mqttc = connect()
    while not exit_flag:
        reconnect_interval = 5

        try:
            mqttc.loop_forever()
        except socket.error as ex:
            logger.exception(f"Connection to MQTT broker lost: {ex.__class__.__name__}({ex})")
        except Exception as ex:
            logger.exception(f"Connection to MQTT broker lost: {ex.__class__.__name__}({ex})")
            raise

        if not exit_flag:
            logger.warning("MQTT server disconnected, trying to reconnect each %s seconds" % reconnect_interval)
            # TODO: Review this.
            time.sleep(reconnect_interval)


def publish_status_information():
    """
    Implement `$SYS` topic, like Mosquitto's "Broker Status".
    https://mosquitto.org/man/mosquitto-8.html#idm289

    Idea from Mosquitto::

      $ mosquitto_sub -t '$SYS/broker/version' -v
      $SYS/broker/version mosquitto version 2.0.10

    Synopsis::

      $ mosquitto_sub -t 'mqttwarn/$SYS/#' -v

      mqttwarn/$SYS/version 0.26.2
      mqttwarn/$SYS/platform darwin
      mqttwarn/$SYS/python/version 3.9.7

    """
    if cf.has_option("defaults", "status_publish") and cf.status_publish:

        status_topic = cf.g("defaults", "status_topic", "mqttwarn/$SYS")
        logger.info(f"Publishing status information to {status_topic}")

        # Items are tuples of (subtopic, message)
        si = StatusInformation()
        publications = [
            ("version", si.mqttwarn_version),
            ("platform", si.os_platform),
            ("python/version", si.python_version),
        ]
        try:
            for publication in publications:
                subtopic, message = publication
                mqttc.publish(status_topic + "/" + subtopic, message, retain=True)

        except:
            logger.exception("Unable to publish status information")


def start_workers():

    # Launch worker threads to operate on queue
    logger.info("Starting %s worker threads" % cf.num_workers)
    for i in range(cf.num_workers):
        t = threading.Thread(target=processor, kwargs={"worker_id": i})
        t.daemon = True
        t.start()

    # If the config file has a [cron] section, the key names therein are
    # functions from 'myfuncs.py' which should be invoked periodically.
    # The key's value (must be numeric!) is the period in seconds.

    if cf.has_section("cron"):
        for name, val in cf.items("cron"):
            try:
                func = load_function(name=name, py_mod=cf.functions)
            except AttributeError:
                logger.error("[cron] section has function [%s] specified, but that's not defined" % name)
                continue

            cron_options = parse_cron_options(val)
            interval = cron_options["interval"]
            logger.info(
                "Scheduling periodic task '{name}' to run each "
                "{interval} seconds via [cron] section".format(name=name, interval=interval)
            )
            service = make_service(mqttc=mqttc, name="mqttwarn.cron")
            ptlist[name] = PeriodicThread(
                callback=func, period=interval, name=name, srv=service, now=asbool(cron_options.get("now"))
            )
            ptlist[name].start()


def cleanup(signum=None, frame=None):
    """
    Signal handler to ensure we disconnect cleanly
    in the event of a SIGTERM or SIGINT.
    """
    for ptname in ptlist:
        logger.info(f"Cancelling periodic task '{ptname}'")
        ptlist[ptname].cancel()

    logger.debug("Disconnecting from MQTT broker")
    if cf.lwt is not None:
        mqttc.publish(cf.lwt, cf.lwt_dead, qos=0, retain=True)
    mqttc.loop_stop()
    mqttc.disconnect()

    logger.info("Waiting for queue to drain")
    q_in.join()

    # Send exit signal to subsystems _after_ queue was drained.
    # TODO: Refactor this elsewhere.
    global exit_flag
    exit_flag = True

    # TODO: Refactor this elsewhere.
    logger.debug(f"Exiting on signal {signum}")
    # TODO: Review this.
    sys.exit(signum)


def bootstrap(config=None, scriptname=None):
    # FIXME: Remove global variables
    global context, cf, SCRIPTNAME
    # NOTE: this is called before we connect to the MQTT broker, so mqttc is not initialised yet
    invoker = FunctionInvoker(config=config, srv=make_service(mqttc=None, name="mqttwarn.context"))
    context = RuntimeContext(config=config, invoker=invoker)
    cf = config
    if scriptname is not None:
        SCRIPTNAME = scriptname


def run_plugin(config=None, name=None, options=None, data=None, message=None):
    """
    Run service plugins directly without the
    dispatching and transformation machinery.

    On the one hand, this might look like a bit of a hack.
    On the other hand, it shows very clearly how some of
    the innards of mqttwarn interact so it might also please
    newcomers as a "learning the guts of mqttwarn" example.

    :param config:   The configuration object
    :param name:     The name of the service plugin, e.g. "log" or "file"
    :param options:  The data to be converged into an appropriate item Struct object instance
    """

    # Bootstrap mqttwarn core
    # TODO: Optionally run w/o configuration.
    bootstrap(config=config)

    # Load designated service plugins
    load_services([name])
    if "." in name:
        service_logger_name = name
    else:
        service_logger_name = "mqttwarn.services.{}".format(name)
    srv = make_service(mqttc=None, name=service_logger_name)

    # Build a mimikry item instance for feeding to the service plugin
    item = Struct(**options or {})
    # TODO: Read configuration optionally from data.
    item.config = config.config("config:" + name)
    item.service = srv
    item.target = "mqttwarn"
    item.data = data or {}
    if not hasattr(item, "message"):
        item.message = message

    # Launch plugin
    module = service_plugins[name]["module"]
    response = module.plugin(srv, item)
    logger.info("Plugin response: {}".format(response))
    if response is False:
        sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Jan-Piet Mens <jpmens()gmail.com>'
__copyright__ = 'Copyright 2014 Jan-Piet Mens'
__license__ = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'

import socket


def plugin(srv, item):
    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    try:
        addr, port, channel = item.addrs
    except:
        srv.logging.warning("Incorrect target configuration")
        return False

    message = item.message

    # Optionally apply coloring.
    color = None
    if item.priority == 1:
        color = '%GREEN'
    elif item.priority == 2:
        color = '%RED'
    if color is not None:
        message = color + message

    srv.logging.debug("Sending to IRCcat: %s" % (message))

    # Apparently, a trailing newline is needed.
    # https://github.com/mqtt-tools/mqttwarn/issues/547#issuecomment-944632712
    message += "\n"

    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((addr, port))
        sock.send(message.encode())
        sock.close()

    except Exception as e:
        srv.logging.error("Error sending IRCcat notification to %s:%s [%s]: %s" % (item.target, addr, port, e))
        return False

    return True



# -*- coding: utf-8 -*-
# (c) 2021-2023 The mqttwarn developers
from __future__ import absolute_import

from functools import lru_cache
from urllib.parse import urlparse, urlencode

from apprise import Apprise, ContentLocation

from mqttwarn.model import ProcessorItem


@lru_cache(maxsize=None)
def get_all_template_argument_names():
    """
    Inquire all possible parameter names from all Apprise plugins.
    """
    a = Apprise(asset=None, location=ContentLocation.LOCAL)
    results = a.details()
    plugin_infos = results['schemas']

    all_arg_names = []
    for plugin_info in plugin_infos:
        arg_names = plugin_info["details"]["args"].keys()
        all_arg_names += arg_names

    return sorted(set(all_arg_names))


def obtain_apprise_arguments(item: ProcessorItem, arg_names: list) -> dict:
    """
    Obtain eventual Apprise parameters from data dictionary.
    """
    params = dict()
    for arg_name in arg_names:
        if isinstance(item.data, dict) and arg_name in item.data:
            params[arg_name] = item.data[arg_name]
    return params


def add_url_params(url: str, params: dict) -> str:
    """
    Serialize query parameter dictionary and add it to URL.
    """
    url_parsed = urlparse(url)
    if params:
        seperator = "?"
        if url_parsed.query:
            seperator = "&"
        url += seperator + urlencode(params)
    return url



# -*- coding: utf-8 -*-
# mqttwarn example function extensions
import time
import copy

try:
    import json
except ImportError:
    import simplejson as json  # type: ignore[no-redef]

def OwnTracksTopic2Data(topic):
    try:
        # owntracks/username/device
        parts = topic.split('/')
        username = parts[1]
        deviceid = parts[2]
    except:
        deviceid = 'unknown'
        username = 'unknown'
    return dict(username=username, device=deviceid)

def OwnTracksConvert(data):
    if type(data) == dict:
        # Better safe than sorry: Clone transformation dictionary to prevent leaking local modifications
        # See also https://github.com/jpmens/mqttwarn/issues/219#issuecomment-271815495
        data = copy.copy(data)
        tst = data.get('tst', int(time.time()))
        data['tst'] = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(int(tst)))
        # Remove these elements to eliminate warnings
        for k in ['_type', 'desc']:
            data.pop(k, None)

        return "{username} {device} {tst} at location {lat},{lon}".format(**data)

# custom function to filter out any OwnTracks notifications which do
# not contain the 'batt' parameter
def OwnTracksBattFilter(topic, message):
    data = dict(json.loads(message).items())
    if 'batt' in data:
        if data['batt'] is not None:
            return int(data['batt']) > 20

    return True     # Suppress message because no 'batt'

def TopicTargetList(topic=None, data=None, srv=None):
    """
    Custom function to compute list of topic targets based on MQTT topic and/or transformation data.
    Obtains MQTT topic, transformation data and service object.
    Returns list of topic target identifiers.
    """

    # optional debug logger
    if srv is not None:
        srv.logging.debug('topic={topic}, data={data}, srv={srv}'.format(**locals()))

    # Use a fixed list of topic targets for demonstration purposes.
    targets = ['log:debug']

    # In the real world, you would compute proper topic targets based on information
    # derived from transformation data, which in turn might have been enriched
    # by ``datamap`` or ``alldata`` transformation functions before, like that:
    if 'condition' in data:

        if data['condition'] == 'sunny':
            targets.append('file:mqttwarn')

        elif data['condition'] == 'rainy':
            targets.append('log:warn')

    return targets

def publish_public_ip_address(srv=None):
    """
    Custom function used as a periodic task for publishing your public ip address to the MQTT bus.
    Obtains service object.
    Returns None.
    """

    import socket
    import requests

    hostname = socket.gethostname()
    ip_address = requests.get('https://httpbin.org/ip').json().get('origin')

    if srv is not None:

        # optional debug logger
        srv.logging.debug('Publishing public ip address "{ip_address}" of host "{hostname}"'.format(**locals()))

        # publish ip address to mqtt
        srv.mqttc.publish('test/ip/{hostname}'.format(**locals()), ip_address)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__    = 'Jan-Piet Mens <jpmens()gmail.com>'
__copyright__ = 'Copyright 2014 Jan-Piet Mens'
__license__   = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'

import io
import tempfile


def plugin(srv, item):

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    mode = "a"

    # item.config is brought in from the configuration file
    config = item.config

    # Evaluate global parameters.
    newline = False
    overwrite = False
    if type(config) == dict and 'append_newline' in config and config['append_newline']:
        newline = True
    if type(config) == dict and 'overwrite' in config and config['overwrite']:
        overwrite = True

    # `item.addrs` is either a dict or a list associated with a particular target.
    # While lists may contain more than one item (e.g., for the pushover target),
    # the `file` service only allows for single items, the path name.
    # When it's a dict, additional parameters can be obtained to augment the
    # behavior of the write operation on a per-file basis.
    if isinstance(item.addrs, dict):
        filename = item.addrs['path'].format(**item.data)
        # Evaluate per-file parameters.
        newline = item.addrs.get('append_newline', newline)
        overwrite = item.addrs.get('overwrite', overwrite)
    else:
        filename = item.addrs[0].format(**item.data)

    # Interpolate some variables into filename.
    if "$TMPDIR" in filename:
        filename = filename.replace("$TMPDIR", tempfile.gettempdir())

    srv.logging.info("Writing to file `%s'" % (filename))

    # If the incoming payload has been transformed, use that,
    # else the original payload
    text = item.message

    if newline:
        text += "\n"
    if overwrite:
        mode = "w"

    if isinstance(text, bytes):
        mode += "b"
        encoding = None
    else:
        encoding = "utf-8"

    try:
        f = io.open(filename, mode=mode, encoding=encoding)
        f.write(text)
        f.close()

    except Exception as e:
        srv.logging.error("Cannot write to file `%s': %s" % (filename, e))
        return False

    return True



import warnings

from mqttwarn.services.apprise_single import plugin

warnings.warn("`mqttwarn.services.apprise` will be removed in a future release of mqttwarn. "
              "Please use `mqttwarn.services.apprise_single` or `mqttwarn.services.apprise_multi` instead.",
              category=DeprecationWarning)



# -*- coding: utf-8 -*-
"""
Frigate » Forward events and snapshots to Ntfy, using mqttwarn.

https://mqttwarn.readthedocs.io/en/latest/examples/frigate/README.html
"""
import dataclasses
import json
import re
import typing as t
from collections import OrderedDict
from datetime import datetime, timezone

from mqttwarn.context import RuntimeContext
from mqttwarn.model import Service


@dataclasses.dataclass
class FrigateEvent:
    """
    Manage inbound event data received from Frigate.
    """

    time: datetime
    camera: str
    label: str
    current_zones: t.List[str]
    entered_zones: t.List[str]

    @staticmethod
    def format_list(value: t.List[str]) -> t.List[str]:
        """
        Format a list for human consumption.
        """
        return [y.replace("_", " ") for y in value]

    @property
    def current_zones_str(self) -> str:
        """
        Serialize list of `current_zones` to string.
        """
        return ", ".join(self.format_list(self.current_zones or []))

    @property
    def entered_zones_str(self) -> str:
        """
        Serialize list of `entered_zones` to string.
        """
        return ", ".join(self.format_list(self.entered_zones or []))

    def to_dict(self) -> t.Dict[str, str]:
        """
        Return Python dictionary from attributes.
        """
        return dataclasses.asdict(self)

    @classmethod
    def from_json(cls, payload: str) -> "FrigateEvent":
        """
        Decode inbound Frigate event, in JSON format.
        """
        # Decode JSON message.
        after = json.loads(payload)["after"]

        # Decode inbound Frigate event.
        return cls(
            time=datetime.fromtimestamp(after["frame_time"], tz=timezone.utc),
            camera=after["camera"],
            label=after["sub_label"] or after["label"],
            current_zones=after["current_zones"],
            entered_zones=after["entered_zones"],
        )


ContainerType = t.Dict[str, t.Union[str, FrigateEvent]]


def frigate_events(topic: str, data: t.Dict[str, str], srv: Service) -> ContainerType:
    """
    mqttwarn transformation function which computes options to be submitted to ntfy.
    """

    # Decode inbound Frigate event.
    event = FrigateEvent.from_json(data["payload"])

    # Collect outbound ntfy option fields.
    params: ContainerType = OrderedDict()
    params.update(event.to_dict())

    # Also add the event object as a whole, to let downstream templates leverage it.
    params["event"] = event

    return params


def frigate_events_filter(topic: str, payload: str, section: str, srv: Service) -> bool:
    """
    mqttwarn filter function to only use `new` and important `update` Frigate events.

    Additionally, validate more details within the event message,
    specifically the `after` section. For example, skip false positives.

    :return: True if message should be filtered, i.e. notification should be skipped.
    """
    try:
        message = json.loads(payload)
    except json.JSONDecodeError as e:
        srv.logging.warning(f"Can't parse Frigate event message: {e}")
        return True

    # ignore ending messages
    message_type = message.get("type")
    if message_type == "end":
        srv.logging.warning(f"Frigate event skipped, ignoring Message type '{message_type}'")
        return True

    # payload must have 'after' key
    elif "after" not in message:
        srv.logging.warning("Frigate event skipped, 'after' missing from payload")
        return True

    after = message.get("after")

    nonempty_fields = ["false_positive", "camera", "label", "current_zones", "entered_zones", "frame_time"]
    for field in nonempty_fields:

        # Validate field exists.
        if field not in after:
            srv.logging.warning(f"Frigate event skipped, missing field: {field}")
            return True

        value = after.get(field)

        # We can ignore if `current_zones` is empty.
        if field == "current_zones":
            continue

        # Check if it's a false positive.
        if field == "false_positive":
            if value is True:
                srv.logging.warning("Frigate event skipped, it is a false positive")
                return True
            else:
                continue

        # All other keys should be present and have values.
        if not value:
            srv.logging.warning(f"Frigate event skipped, field is empty: {field}")
            return True

    # Ignore unimportant `update` events.
    before = message.get("before")
    if message_type == "update" and isinstance(before, dict):
        if before.get("stationary") is True and after.get("stationary") is True:
            srv.logging.warning("Frigate event skipped, object is stationary")
            return True
        elif after["current_zones"] == after["entered_zones"] or (
            before["current_zones"] == after["current_zones"] and before["entered_zones"] == after["entered_zones"]
        ):
            srv.logging.warning("Frigate event skipped, object stayed within same zone")
            return True

    # Evaluate optional skip rules.
    context: RuntimeContext = srv.mwcore["context"]
    frigate_skip_rules = context.config.getdict(section, "frigate_skip_rules")
    for rule in frigate_skip_rules.values():
        do_skip = True
        for field_name, skip_values in rule.items():
            actual_value = after[field_name]
            if isinstance(actual_value, list):
                do_skip = do_skip and all(value in skip_values for value in actual_value)
            else:
                do_skip = do_skip and actual_value in skip_values
        if do_skip:
            srv.logging.warning("Frigate event skipped, object did not enter zone of interest")
            return True

    return False


def frigate_snapshot_decode_topic(topic: str, data: t.Dict[str, str], srv: Service) -> t.Optional[t.Dict[str, str]]:
    """
    Decode Frigate MQTT topic for image snapshots.

    frigate/+/+/snapshot

    See also:
    - https://docs.frigate.video/integrations/mqtt/#frigatecamera_nameobject_namesnapshot
    """
    topology = {}
    if isinstance(topic, str):
        try:
            # TODO: Compile pattern only once, for efficiency reasons.
            pattern = r"^frigate/(?P<camera_name>.+?)/(?P<object_name>.+?)/snapshot$"
            p = re.compile(pattern)
            m = p.match(topic)
            if m:
                topology = m.groupdict()
        except:
            pass
    return topology



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'hubble2webb <hubble2webb@users.noreply.github.com>'
__copyright__ = 'Copyright 2015 hubble2webb'
__license__ = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'

from future import standard_library
standard_library.install_aliases()
from builtins import str

import urllib.request, urllib.error, urllib.parse
import base64

try:
    import simplejson as json
except ImportError:
    import json  # type: ignore[no-redef]


def plugin(srv, item):
    # item.addrs is an array with three or more elements:
    # 0 is the Ionic appid
    # 1 is the Ionic appsecret (private key)
    # 2..N are the push tokens returned by Ionic push service

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    if len(item.addrs) < 3:
        srv.logging.error("appid, appsecret and atleast one devicetoken is required")
        return False

    appid = item.addrs[0]
    appsecret = item.addrs[1]
    devicetokens = item.addrs[2:]

    if not appid or appid.isspace():
        srv.logging.error("appid is missing or empty")
        return False
    if not appsecret or appsecret.isspace():
        srv.logging.error("appsecret is missing or empty")
        return False
    if len(devicetokens) == 0:
        srv.logging.error("atleast one devicetoken is required")
        return False

    devicetokens = [_f for _f in devicetokens if _f]
    devicetokens = [name for name in devicetokens if name.strip()]
    appid = appid.strip()
    appsecret = appsecret.strip()


    data = {"tokens": devicetokens}
    notification = {"alert": item.message}
    data["notification"] = notification

    resource = "https://push.ionic.io/api/v1/push"

    try:
        handler = urllib.request.HTTPHandler()
        opener = urllib.request.build_opener(handler)

        credentials = '%s:' % (appsecret)
        basicauth_token = base64.b64encode(credentials.encode('utf-8')).decode()

        data = json.dumps(data)
        request = urllib.request.Request(resource, data=data.encode("utf-8"))
        request.add_header('X-Ionic-Application-Id', appid)
        request.add_header("Authorization", "Basic %s" % basicauth_token)
        request.add_header("Content-Type", 'application/json')

        connection = opener.open(request, timeout=5)
        srv.logging.info("Server reply: %s" % str(connection.read()))

    except urllib.error.HTTPError as e:
        srv.logging.warn("Failed to send POST request to ionic using %s: %s" % (resource, str(e.read())))
        return False

    return True



#!/usr/bin/python

import json

def index(req):

    filename = "/var/www/html/warntoggle/warntoggle.json"
    writetoggles = {}
    readtoggles = {} 
    scriptname = __file__.split('/')[-1:][0]

    # write form data to the file if applicable

    if bool(req.form): 

        for k,v in req.form.items():
            writetoggles[k] = eval(str(v))

        with open(filename, 'w') as outfile: 
            json.dump(writetoggles, outfile)
            outfile.close()

    # read data from the file

    with open(filename, 'r') as infile:
        readtoggles = json.load(infile)
        infile.close()

    # display the form

    html  = "<html><body><h1>mqttwarn notification toggles</h1>"
    html += "<table>"
    html += "<form action=\"" + scriptname + "\" method=\"post\">"

    for key in sorted(readtoggles.keys()):
        if readtoggles[key] == True:
            block_checked = " checked"
            notify_checked = ""
        else:
            block_checked = ""
            notify_checked = " checked"

        html += "<tr><td>" + key + "</td><td><input type='radio' name='" + key + "' value='True'" + block_checked + ">Block <input type='radio' name='" + key + "' value='False'" + notify_checked+ ">Notify</td></tr>"

    html += "</table><br/>"
    html += "<nobr><input type='submit' value='Submit'> <a href='" + scriptname + "'>reload</a></nobr>"
    html += "</form>"
    html += "</body></html>"

    return html 



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__    = 'Jan-Piet Mens <jpmens()gmail.com>'
__copyright__ = 'Copyright 2014 Jan-Piet Mens'
__license__   = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'

import subprocess


def plugin(srv, item):

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    voice = item.addrs[0]
    text = item.message

    argv = [ "/usr/bin/say", "-f", "-", "--voice=%s" % voice ]

    try:
        proc = subprocess.Popen(argv,
            stdin=subprocess.PIPE, close_fds=True)
    except Exception as e:
        srv.logging.warn("Cannot create osxsay pipe: %s" % e)
        return False

    try:
        proc.stdin.write(text)
    except IOError as e:
        srv.logging.warn("Cannot write to osxsay pipe: errno %d" % (e.errno))
        return False
    except Exception as e:
        srv.logging.warn("Cannot write to osxsay pipe: %s" % e)
        return False

    proc.stdin.close()
    proc.wait()
    return True



# -*- coding: utf-8 -*-
# Homie function extensions for mqttwarn
import re


# ------------------------------------------
#                  Synopsis
# ------------------------------------------
#
# Run mqttwarn::
#
#   export MQTTWARNINI=examples/homie/homie.ini
#   mqttwarn
#
# Send some homie-like data::
#
#   mosquitto_pub -t homie/bee1/weight/value -m 42.42


def decode_homie_topic(topic):
    """
    Split Homie-style MQTT topic path into segments for
    enriching transformation data inside mqttwarn.
    """
    if type(topic) == str:
        try:
            pattern = r'^(?P<realm>.+?)/(?P<device>.+?)/(?P<node>.+?)/(?P<property>.+?)$'
            p = re.compile(pattern)
            m = p.match(topic)
            topology = m.groupdict()
        except:
            topology = {}
        return topology
    return None




#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__    = 'Fabian Affolter <fabian()affolter-engineering.ch>'
__copyright__ = 'Copyright 2014 Fabian Affolter'
__license__   = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'

import dbus


def plugin(srv, item):
    """Send a message through dbus to the user's desktop."""

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    text = item.message
    summary = item.addrs[0]
    app_name = item.get('title', srv.SCRIPTNAME)
    replaces_id = 0
    service = 'org.freedesktop.Notifications'
    path = '/' + service.replace('.', '/')
    interface = service
    app_icon = '/usr/share/icons/gnome/32x32/places/network-server.png'
    expire_timeout = 1000
    actions = []
    hints = []

    try:
        srv.logging.debug("Sending message to %s..." % (item.target))
        session_bus = dbus.SessionBus()
        obj = session_bus.get_object(service, path)
        interface = dbus.Interface(obj, interface)
        interface.Notify(app_name, replaces_id, app_icon, summary, text,
                    actions, hints, expire_timeout)
        srv.logging.debug("Successfully sent message")
    except Exception as e:
        srv.logging.error("Error sending message to %s: %s" % (item.target, e))
        return False

    return True



# -*- coding: utf-8 -*-
# (c) 2014-2023 The mqttwarn developers
import logging
import threading
import typing as t

from mqttwarn.model import Service

logger = logging.getLogger(__name__)


# This class, shamelessly stolen from https://gist.github.com/cypreess/5481681
# The `srv` bits are added for mqttwarn
class PeriodicThread:
    """
    Python periodic Thread using Timer with instant cancellation
    """

    def __init__(
        self,
        callback: t.Optional[t.Callable] = None,
        period: t.Optional[int] = 1,
        name: t.Optional[str] = None,
        srv: t.Optional[Service] = None,
        now: t.Optional[bool] = False,
        *args,
        **kwargs,
    ):
        self.name = name
        self.srv = srv
        self.now = now
        self.args = args
        self.kwargs = kwargs
        self.callback = callback
        self.period = period
        self.stop = False
        self.current_timer = None
        self.schedule_lock = threading.Lock()

    def start(self):
        """
        Mimics Thread standard start method
        """

        # Schedule periodic task to run right now
        if self.now is True:
            self._run()

        # Schedule periodic task with designated interval
        else:
            self.schedule_timer()

    def run(self):
        """
        By default run callback. Override it if you want to use inheritance
        """
        if self.callback is not None:
            self.callback(self.srv, *self.args, **self.kwargs)

    def _run(self):
        """
        Run desired callback and then reschedule Timer (if thread is not stopped)
        """
        try:
            self.run()
        except Exception:
            logger.exception(f"Exception while running periodic thread '{self.name}'")
        finally:
            with self.schedule_lock:
                if not self.stop:
                    self.schedule_timer()

    def schedule_timer(self):
        """
        Schedules next Timer run
        """
        self.current_timer = threading.Timer(self.period, self._run)
        if self.name:
            self.current_timer.name = self.name
        self.current_timer.start()

    def cancel(self):
        """
        Mimics Timer standard cancel method
        """
        with self.schedule_lock:
            self.stop = True
            if self.current_timer is not None:
                self.current_timer.cancel()

    def join(self):
        """
        Mimics Thread standard join method
        """
        self.current_timer.join()



"""
Forward OwnTracks low-battery warnings to ntfy.
https://mqttwarn.readthedocs.io/en/latest/examples/owntracks-ntfy/readme.html
"""
import json


def owntracks_batteryfilter(topic: str, message: str):
    ignore = True
    try:
        data = dict(json.loads(message).items())
    except:
        data = None

    if data and "batt" in data and data["batt"] is not None:
        ignore = int(data["batt"]) > 20

    return ignore



# -*- coding: utf-8 -*-
# (c) 2023 The mqttwarn developers
"""
Software test accompanying "Forwarding Frigate events to ntfy" example.
Full system test verifying mqttwarn end-to-end, using an MQTT broker, and an ntfy instance.
"""
import os
import subprocess
import sys
from pathlib import Path

import pytest
import requests

from mqttwarn.configuration import load_configuration
from mqttwarn.core import bootstrap, connect
from tests.util import mqtt_process

if os.getenv("GITHUB_ACTIONS") == "true" and sys.platform != "linux":
    raise pytest.skip(msg="On GHA, Mosquitto via Docker is only available on Linux", allow_module_level=True)


# Mark all test case functions within this module with `e2e`.
pytestmark = pytest.mark.e2e

# Configure `capmqtt` to not decode `MqttMessage.payload` from `utf-8`, to make it possible to process binary data.
capmqtt_decode_utf8 = False

HERE = Path(__file__).parent
ASSETS = HERE.joinpath("assets")


@pytest.fixture(scope="function", autouse=True)
def reset_filestore():
    """
    Make sure there are no temporary spool files around before starting each test case.
    """
    for filepath in Path("/tmp").glob("mqttwarn-frigate-*"):
        os.unlink(filepath)


def test_frigate_with_attachment(mosquitto, ntfy_service, caplog, capmqtt):
    """
    A full system test verifying the "Forwarding Frigate events to ntfy" example.
    This test case submits the `frigate-event-full.json` file as MQTT event message.
    """

    # Bootstrap the core machinery without MQTT.
    config = load_configuration(configfile=HERE / "frigate.ini")
    bootstrap(config=config)

    # Get image payload.
    payload_image = get_goat_image()

    # Add MQTT.
    mqttc = connect()

    # Make mqttwarn run the subscription to the broker.
    mqtt_process(mqttc)

    # Publish the JSON event message.
    # cat frigate-event-new-good.json | jq -c | mosquitto_pub -t 'frigate/events' -l
    payload_event = open(ASSETS / "frigate-event-full.json").read()
    capmqtt.publish(topic="frigate/events", payload=payload_event)

    # Publish the snapshot image.
    # mosquitto_pub -f goat.png -t 'frigate/cam-testdrive/goat/snapshot'
    capmqtt.publish(topic="frigate/cam-testdrive/goat/snapshot", payload=payload_image)

    # Make mqttwarn receive and process the message.
    mqtt_process(mqttc, loops=6)

    # Verify log output.
    assert 'Successfully loaded service "ntfy"' in caplog.messages
    assert 'Successfully loaded service "store-image"' in caplog.messages

    assert "Subscribing to frigate/events (qos=0)" in caplog.messages
    assert "Subscribing to frigate/+/+/snapshot (qos=0)" in caplog.messages

    assert "MQTT message received: MqttMessage(topic='frigate/events'" in caplog.text
    assert "Section [frigate/events] matches message on frigate/events, processing it" in caplog.messages
    assert "Message on frigate/events going to ntfy:test" in caplog.messages
    assert "Invoking service plugin for `ntfy'" in caplog.messages
    assert (
        "Headers: {"
        "'Click': 'https://httpbin.org/anything?camera=cam-testdrive&label=goat&zone=lawn', "
        "'Title': '=?utf-8?q?goat_entered_lawn_at_2023-04-06_14=3A31=3A46=2E638857+00=3A00?=', "
        "'Message': '=?utf-8?q?goat_was_in_barn_before?=', "
        "'Filename': 'mqttwarn-frigate-cam-testdrive-goat.png'}" in caplog.messages
    )
    assert (
        "Sending notification to ntfy. target=test, options={"
        "'url': 'http://username:password@localhost:5555/frigate-testdrive', "
        "'file': '/tmp/mqttwarn-frigate-{camera}-{label}.png', "
        "'click': 'https://httpbin.org/anything?camera={event.camera}&label={event.label}&zone={event.entered_zones[0]}'}"
        in caplog.messages
    )

    # assert "Sent ntfy notification to 'http://localhost:5555'." in caplog.messages
    assert "Successfully sent message using ntfy" in caplog.messages

    assert "MQTT message received: MqttMessage(topic='frigate/cam-testdrive/goat/snapshot'" in caplog.text
    assert (
        "Section [frigate/+/+/snapshot] matches message on frigate/cam-testdrive/goat/snapshot, processing it"
        in caplog.messages
    )
    assert "Message on frigate/cam-testdrive/goat/snapshot going to store-image:cam-testdrive-goat" in caplog.messages
    assert "Writing to file `/tmp/mqttwarn-frigate-cam-testdrive-goat.png'" in caplog.messages

    assert "Job queue has 0 items to process" in caplog.messages
    mqttc.disconnect()

    # Verify notification was properly received by ntfy.
    url = "http://localhost:5555/frigate-testdrive/json?poll=1&since=5s"
    response = requests.get(url)
    data = response.json()
    del data["id"]
    del data["time"]
    del data["expires"]
    del data["attachment"]["url"]
    del data["attachment"]["expires"]
    assert data == {
        "event": "message",
        "topic": "frigate-testdrive",
        "title": "goat entered lawn at 2023-04-06 14:31:46.638857+00:00",
        "message": "goat was in barn before",
        "click": "https://httpbin.org/anything?camera=cam-testdrive&label=goat&zone=lawn",
        "attachment": {
           "name": "mqttwarn-frigate-cam-testdrive-goat.png",
           "size": 283595,
           "type": "image/png",
        },
    }


@pytest.mark.parametrize(
    "jsonfile", ["frigate-event-full.json", "frigate-event-new-good.json", "frigate-event-update-good.json"]
)
def test_frigate_with_notification(mosquitto, ntfy_service, caplog, capmqtt, jsonfile):
    """
    A full system test verifying the "Forwarding Frigate events to ntfy" example.
    This test case submits JSON files which trigger a notification.
    """

    # Bootstrap the core machinery without MQTT.
    config = load_configuration(configfile=HERE / "frigate.ini")
    turn_off_retries(config)
    bootstrap(config=config)

    # Add MQTT.
    mqttc = connect()

    # Make mqttwarn run the subscription to the broker.
    mqtt_process(mqttc)

    # Publish the JSON event message.
    # cat frigate-event-full.json | jq -c | mosquitto_pub -t 'frigate/events' -l
    payload_event = open(ASSETS / jsonfile).read()
    capmqtt.publish(topic="frigate/events", payload=payload_event)

    # Make mqttwarn receive and process the message.
    mqtt_process(mqttc, loops=3)

    # Verify log output.
    assert "MQTT message received: MqttMessage(topic='frigate/events'" in caplog.text
    assert "Section [frigate/events] matches message on frigate/events, processing it" in caplog.messages
    assert "Message on frigate/events going to ntfy:test" in caplog.messages
    assert "Invoking service plugin for `ntfy'" in caplog.messages
    assert (
        "Headers: {"
        "'Click': 'https://httpbin.org/anything?camera=cam-testdrive&label=goat&zone=lawn', "
        "'Title': '=?utf-8?q?goat_entered_lawn_at_2023-04-06_14=3A31=3A46=2E638857+00=3A00?=', "
        "'Message': '=?utf-8?q?goat_was_in_barn_before?='}" in caplog.messages
    )

    # assert "Sent ntfy notification to 'http://localhost:5555'." in caplog.messages
    assert "Successfully sent message using ntfy" in caplog.messages

    assert "Job queue has 0 items to process" in caplog.messages
    mqttc.disconnect()


@pytest.mark.parametrize(
    "jsonfile",
    [
        "frigate-event-end.json",
        "frigate-event-false-positive.json",
        "frigate-event-update-samezone.json",
        "frigate-event-update-stationary.json",
        "frigate-event-new-ignored.json",
    ],
)
def test_frigate_without_notification(mosquitto, caplog, capmqtt, jsonfile):
    """
    A full system test verifying the "Forwarding Frigate events to ntfy" example.
    This test case submits JSON files which should not trigger a notification.
    """

    # Bootstrap the core machinery without MQTT.
    config = load_configuration(configfile=HERE / "frigate.ini")
    bootstrap(config=config)

    # Add MQTT.
    mqttc = connect()

    # Make mqttwarn run the subscription to the broker.
    mqtt_process(mqttc)

    # Publish the JSON event message.
    # cat frigate-event-full.json | jq -c | mosquitto_pub -t 'frigate/events' -l
    payload_event = open(ASSETS / jsonfile).read()
    capmqtt.publish(topic="frigate/events", payload=payload_event)

    # Make mqttwarn receive and process the message.
    mqtt_process(mqttc, loops=3)

    # Verify log output.
    assert "MQTT message received: MqttMessage(topic='frigate/events'" in caplog.text
    assert "Section [frigate/events] matches message on frigate/events, processing it" in caplog.messages
    assert "Frigate event skipped" in caplog.text
    assert "Filter in section [frigate/events] has skipped message on frigate/events" in caplog.messages

    assert "Job queue has 0 items to process" in caplog.messages
    mqttc.disconnect()


def get_goat_image() -> bytes:
    """
    Get an image of a Changthangi goat.
    """
    return requests.get(
        "https://user-images.githubusercontent.com/453543/231550862-5a64ac7c-bdfa-4509-86b8-b1a770899647.png"
    ).content


def turn_off_retries(config):
    """
    For test cases which do not aim to submit any attachments, adjust the configuration
    to not wait for it to appear. Otherwise, it would take too long, and the verification
    would fail.
    """
    targets = config.getdict("config:ntfy", "targets")
    targets["test"]["__settings__"]["file_retry_tries"] = None
    config.set("config:ntfy", "targets", targets)



# -*- coding: utf-8 -*-
# (c) 2014-2023 The mqttwarn developers
import os
import platform
import sys

from setuptools import find_packages, setup
from versioningit import get_cmdclasses

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, "README.rst")).read()

requires = [
    "attrs<23",
    "docopt<1",
    "funcy<3",
    "future>=0.18.0,<1",
    "importlib-metadata; python_version<'3.8'",
    "jinja2<4",
    "paho-mqtt<2",
    "requests<3",
    "six<2",
]

extras = {
    "amqp": [
        "puka>=0.0.7",
    ],
    "apns": [
        "apns>=2.0.1",
    ],
    "apprise": [
        "apprise>=1.3,<2",
    ],
    "asterisk": [
        "pyst2>=0.5.0",
    ],
    "celery": [
        "celery",
    ],
    "chromecast": [
        "pychromecast>=7.5.0",
    ],
    "dnsupdate": [
        "dnspython>=1.15.0",
    ],
    "fbchat": [
        "fbchat>=1.3.6",
    ],
    "gss2": [
        "google-api-python-client<2",
        "gspread>=2.1.1",
        "oauth2client>=4.1.2",
    ],
    "mysql": [
        "mysql",
    ],
    "mysql_dynamic": [
        "mysqlclient",
    ],
    "nsca": [
        "pynsca>=1.6",
    ],
    "desktopnotify": [
        "desktop-notifier<4",
    ],
    "pastebinpub": [
        "Pastebin>=1.1.2",
    ],
    "postgres": [
        "psycopg2-binary>=2.7.4",
    ],
    "prowl": [
        "pyprowl>=3.0.1",
    ],
    "pushbullet": [
        "PushbulletPythonLibrary>=2.3",
    ],
    "redispub": [
        "redis>=2.10.6",
    ],
    "rrdtool": [
        "rrdtool>=0.1.12",
    ],
    "serial": [
        "pyserial>=3.4",
    ],
    "slack": [
        "slack-sdk>=3.1.0",
    ],
    "ssh": [
        "paramiko>=2.4.1",
    ],
    "tootpaste": [
        "Mastodon.py>=1.2.2",
    ],
    "twilio": [
        "twilio>=6.11.0",
    ],
    "twitter": [
        "python-twitter>=3.4.1",
    ],
    "websocket": [
        "websocket-client>=0.47.0",
    ],
    "xmpp": [
        "xmpppy>=0.6.1",
        "dnspython>=1.16.0",
    ],
    "slixmpp": [
        "slixmpp>=1.5.2",
    ],
    # More notification plugins from the community.
    # https://github.com/daq-tools/mqttwarn-contrib
    "contrib": [
        "mqttwarn-contrib",
    ],
}


# Convenience extra to install *all* dependencies for building a `mqttwarn-full` distribution.
extras_all = []
for extra, packages in extras.items():

    # FIXME: Skip specific packages having build issues.
    # https://github.com/commx/python-rrdtool/issues/36
    if extra in ["mysql", "rrdtool"]:
        continue

    # FIXME: `mysqlclient` needs MySQL or MariaDB client libraries.
    if extra in ["mysql_dynamic"] and sys.platform in ["darwin", "win32"]:
        continue

    # FIXME: Skip specific packages on specific platforms,
    #        because they would also need a build toolchain.
    machine = platform.uname()[4]
    if machine in ["armv7l", "aarch64"] and extra in ["postgres", "slixmpp", "ssh"]:
        continue

    # FIXME: A few packages are not available as wheels on arm32v7, and take quite an amount
    #        of time to build, so let's mask them to improve build times significantly.
    #        Examples: `cryptography`, `aiohttp`, `frozenlist`, `multidict`, `yarl`.
    if machine in ["armv7l"] and extra in ["apprise", "twilio"]:
        continue

    for package in packages:
        extras_all.append(package)

extras["all"] = extras_all


# Packages needed for running the tests.
extras["test"] = [
    "pytest<8",
    "pytest-cov<5",
    "pytest-mock<4",
    "pytest-mqtt<1",
    "tox<4",
    'dataclasses; python_version<"3.7"',
    "requests-toolbelt>=1,<2",
    "responses>=0.13.3,<1",
    "pyfakefs>=4.5,<5",
] + extras["all"]

# Packages needed for development and running CI.
extras["develop"] = [
    "isort<6",
    "black<23",
    "build<1",
    "mypy<1.3",
    "poethepoet<1",
    'ruff==0.0.254; python_version>="3.7"',
    "sphinx-autobuild",
]


setup(
    cmdclass=get_cmdclasses(),
    name="mqttwarn",
    description="mqttwarn - subscribe to MQTT topics and notify pluggable services",
    long_description=README,
    license="EPL 2.0",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Communications",
        "Topic :: Education",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: XMPP",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Pre-processors",
        "Topic :: Software Development :: Testing",
        "Topic :: System :: Archiving",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Monitoring",
        "Topic :: System :: Networking :: Monitoring",
        "Topic :: System :: Systems Administration",
        "Topic :: Text Processing",
        "Topic :: Utilities",
    ],
    author="Jan-Piet Mens, Ben Jones, Andreas Motl",
    author_email="jpmens@gmail.com",
    url="https://github.com/mqtt-tools/mqttwarn",
    keywords="mqtt notification plugins data acquisition push transformation engine mosquitto",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "mqttwarn": [
            "*.ini",
        ],
    },
    zip_safe=False,
    test_suite="tests",
    install_requires=requires,
    extras_require=extras,
    tests_require=extras["test"],
    entry_points={
        "console_scripts": [
            "mqttwarn = mqttwarn.commands:run",
        ],
    },
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Jan-Piet Mens <jpmens()gmail.com>'
__copyright__ = 'Copyright 2014 Jan-Piet Mens'
__license__ = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'

import twitter


def plugin(srv, item):
    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    twitter_keys = item.addrs

    twapi = twitter.Api(
        consumer_key=twitter_keys[0],
        consumer_secret=twitter_keys[1],
        access_token_key=twitter_keys[2],
        access_token_secret=twitter_keys[3]
    )

    text = item.message[0:138]
    try:
        srv.logging.debug("Sending tweet to %s..." % (item.target))
        res = twapi.PostUpdate(text, trim_user=False)
        srv.logging.debug("Successfully sent tweet")
    except twitter.TwitterError as e:
        srv.logging.error("TwitterError: %s" % e)
        return False
    except Exception as e:
        srv.logging.error("Error sending tweet to %s: %s" % (item.target, e))
        return False

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__    = 'Ben Jones <ben.jones12()gmail.com>'
__copyright__ = 'Copyright 2016 Ben Jones'
__license__   = 'Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)'

import requests
import logging

# disable info logging in requests module (e.g. connection pool message for every post request)
logging.getLogger("requests").setLevel(logging.WARNING)


def plugin(srv, item):
    ''' addrs: (measurement) '''

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    host        = item.config['host']
    port        = item.config['port']
    username    = item.config['username']
    password    = item.config['password']
    database    = item.config['database']

    measurement = item.addrs[0]
    tag         = "topic=" + item.topic.replace('/', '_')
    value       = item.message

    # retention policy - "&rp=" is valid in the url, so default=''
    rp          = item.config.get('rp', '')
    # precision=[ns,u,ms,s,m,h] - optional, default=nanosecond
    precision   = item.config.get('precision', 'ns')
    # protocol=[http,https] - optional, default=http
    protocol    = item.config.get('scheme', 'http')

    # allow overrides per target
    # 'target'  = [ 'measurement', 'database',    'rp', 'precision' ]
    if (len(item.addrs) > 1):
        database = item.addrs[1] or database
        if (len(item.addrs) > 2):
            rp = item.addrs[2] or rp
            if (len(item.addrs) > 3):
                precision = item.addrs[3] or precision

    try:
        url = "%s://%s:%d/write?db=%s&rp=%s&precision=%s" % (protocol, host, port, database, rp, precision)

        # influxdb line protocol:
        # measurement,tagKey1=tagVal1,tagKey2=tagVal2 field1=value1,field2=value2 Timestamp
        # sample format in .ini file; no quotes

        if (item.message == item.payload) or (not (' ' in  item.message) and not (',' in item.message)):
            # if no format has been set, default to "value={payload}"
            # or if format has been set to output simple value, without additional tags or multiple fields
            # format = {json_attribute}
            data = measurement + ',' + tag + ' value=' + value

        elif (',' in item.message) and (not ' ' in item.message):
            # if format does not include any additional tags or a timestamp, but includes one or multiple non-default fields
            # format = field1=value1,field2=value2
            data = measurement + ',' + tag + ' ' + item.message

        else:
            # if format includes additional tags and one or multiple non-default fields and an optional timestamp, either group separated by whitespace from each other
            # format = tagKey1=tagVal1,tagKey2=tagVal2 field1=value1,field2=value2
            data = measurement + ',' + tag + ',' + item.message

        srv.logging.debug(url)
        srv.logging.debug(data)

        if username is None:
            r = requests.post(url, data=data)
        else:
            r = requests.post(url, data=data, auth=(username, password))
        
        # success
        if r.status_code == 204:
            return True
            
        # request accepted but couldn't be completed (200) or failed (otherwise)
        if r.status_code == 200:
            srv.logging.warn("POST request could not be completed: %s" % (r.text))
        else:
            srv.logging.warn("POST request failed: (%s) %s" % (r.status_code, r.text))
        
    except Exception as e:
        srv.logging.warn("Failed to send POST request to InfluxDB server using %s: %s" % (url, e))

    return False



__author__ = "Andreas Motl <andreas.motl@panodata.org>"
__copyright__ = "Copyright 2023 Andreas Motl"
__license__ = "Eclipse Public License - v 1.0 (http://www.eclipse.org/legal/epl-v10.html)"

import dataclasses
import logging
from collections import OrderedDict
import typing as t
from email.header import Header
from pathlib import Path

import requests
from funcy import project, merge

from mqttwarn.model import Service, ProcessorItem
from mqttwarn.util import Formatter, asbool, load_file

DataDict = t.Dict[str, t.Union[str, bytes]]


# Field names to be propagated from transformation data to ntfy API.
#
# `topic` will be omitted, and not picked from the transformation
# data, because it contains the MQTT topic already, and would cause
# collisions. The topic is exclusively defined using the `url` field,
# see https://mqttwarn.readthedocs.io/en/latest/notifier-catalog.html#ntfy.
#
# All other ntfy fields are enumerated here.
# https://docs.ntfy.sh/publish/#publish-as-json
# https://docs.ntfy.sh/publish/#list-of-all-parameters

NTFY_FIELD_NAMES: t.List[str] = [
    # "topic",
    "message",
    "title",
    "tags",
    "priority",
    "actions",
    "click",
    "attach",
    "filename",
    "delay",
    "icon",
    "email",
    "cache",
    "firebase",
    "unifiedpush",
]

NTFY_RFC2047_FIELDS: t.List[str] = [
    "message",
    "title",
    "tags",
]

logger = logging.getLogger(__name__)


# The `requests` session instance, for running HTTP requests.
http = requests.Session()
# TODO: Add mqttwarn version.
http.headers.update({"User-Agent": "mqttwarn"})


@dataclasses.dataclass
class NtfyRequest:
    """
    Manage parameters to be propagated to the ntfy HTTP API.
    """

    url: str
    options: t.Dict[str, str]
    fields: DataDict
    attachment_path: t.Optional[str]
    attachment_data: t.Optional[t.Union[bytes, t.IO]]

    def to_http_headers(self) -> t.Dict[str, str]:
        """
        Provide a variant for `fields` to be submitted as HTTP headers to the ntfy API.

        Python's `http.client` will, according to the HTTP specification,
        encode header values using the `latin-1` character set.

        In this spirit, the header transport does not permit any fancy UTF-8 characters
        within any field, so they will be replaced with placeholder characters `?`.
        """
        return dict_with_titles(encode_ntfy_fields(self.fields))


def plugin(srv: Service, item: ProcessorItem) -> bool:
    """
    mqttwarn service plugin for ntfy.
    """

    srv.logging.debug("*** MODULE=%s: service=%s, target=%s", __file__, item.service, item.target)

    # Decode inbound mqttwarn job item into `NtfyRequest`.
    ntfy_request = decode_jobitem(item)

    # Convert field dictionary to HTTP header dictionary.
    headers = ntfy_request.to_http_headers()
    srv.logging.debug(f"Headers: {dict(headers)}")

    # Submit request to ntfy HTTP API.
    try:
        srv.logging.info("Sending notification to ntfy. target=%s, options=%s", item.target, ntfy_request.options)
        response = http.put(ntfy_request.url, data=ntfy_request.attachment_data, headers=headers)
        response.raise_for_status()
    except Exception:
        srv.logging.exception("Request to ntfy API failed")
        return False

    # Report about ntfy response.
    srv.logging.debug(f"ntfy response status: {response}")
    srv.logging.debug(f"ntfy response content: {response.content!r}")

    srv.logging.info("Successfully sent message using ntfy")

    return True


def decode_jobitem(item: ProcessorItem) -> NtfyRequest:
    """
    Decode inbound mqttwarn job item into `NtfyRequest`.
    """

    title = item.title
    body = item.message
    options: t.Dict[str, str]
    settings: t.Dict[str, t.Union[str, int, float, bool]] = {}

    if isinstance(item.addrs, str):
        item.addrs = {"url": item.addrs}
    elif isinstance(item.addrs, dict):
        pass
    else:
        raise TypeError(f"Unable to handle `targets` address descriptor data type `{type(item.addrs).__name__}`: {item.addrs}")

    # Decode options from target address descriptor.
    options = item.addrs

    url = options["url"]
    attachment_path = options.get("file")

    # Extract settings, purging them from the target address descriptor afterwards.
    if "__settings__" in options:
        if isinstance(options["__settings__"], t.Dict):
            settings = dict(options["__settings__"])
            del options["__settings__"]

    # Collect ntfy fields.
    fields: DataDict = OrderedDict()

    # Obtain and propagate all possible ntfy fields from transformation data.
    fields.update(obtain_ntfy_fields(item))

    # Overwrite title and message explicitly, when not present already.
    title and fields.setdefault("title", title)
    body and fields.setdefault("message", body)

    # Attach a file, or not.
    attachment_data = None
    if attachment_path:
        try:
            attachment_path = attachment_path.format(**item.data or {})
            try:
                attachment_data = load_file(
                    path=attachment_path,
                    retry_tries=settings.get("file_retry_tries"),
                    retry_interval=settings.get("file_retry_interval"),
                    unlink=asbool(settings.get("file_unlink")),
                )
                if attachment_data:
                    # TODO: Optionally derive attachment file name from title, using `slugify(title)`.
                    fields.setdefault("filename", Path(attachment_path).name)
            except Exception as ex:
                logger.exception(f"ntfy: Attaching local file failed. Reason: {ex}")
        except:
            logger.exception("ntfy: Computing attachment file name failed")

    ntfy_request = NtfyRequest(
        url=url,
        options=options,
        fields=fields,
        attachment_path=attachment_path,
        attachment_data=attachment_data,
    )

    return ntfy_request


def obtain_ntfy_fields(item: ProcessorItem) -> DataDict:
    """
    Obtain eventual ntfy fields from different sources.
    """

    # Get relevant ntfy option fields from all of `config`,
    # `addrs`, and `data`, with ascending precedence.
    fields_data = item.data and project(item.data, NTFY_FIELD_NAMES) or {}
    fields_addrs = item.addrs and project(item.addrs, NTFY_FIELD_NAMES) or {}
    fields_config = item.config and project(item.config, NTFY_FIELD_NAMES) or {}
    fields: DataDict = OrderedDict()
    fields.update(fields_config)
    fields.update(fields_addrs)
    fields.update(fields_data)

    # Run an interpolation step also on the outbound ntfy option
    # fields, in order to unlock using templated values there.
    logger.info((item.config or {}, item.addrs or {}, item.data or {}))
    all_data = merge(item.config or {}, item.addrs or {}, item.data or {})
    formatter = Formatter()
    for key, value in fields.items():
        if isinstance(value, str):
            new_value = formatter.format(value, **all_data)
            fields[key] = new_value

    return fields


def ascii_clean(data: t.Union[str, bytes]) -> str:
    """
    Return ASCII-clean variant of input string.
    https://stackoverflow.com/a/18430817
    """
    if isinstance(data, bytes):
        data = data.decode()
    if isinstance(data, str):
        return data.encode("ascii", errors="replace").decode()
    else:
        raise TypeError(f"Unknown data type to compute ASCII-clean variant: {type(data).__name__}")


def encode_rfc2047(data: t.Union[str, bytes]) -> str:
    """
    Return RFC2047-encoded variant of input string.

    https://docs.python.org/3/library/email.header.html
    """
    if isinstance(data, bytes):
        data = data.decode()
    if isinstance(data, str):
        return Header(s=data, charset="utf-8").encode()
    else:
        raise TypeError(f"Unknown data type to compute ASCII-clean variant: {type(data).__name__}")


def dict_ascii_clean(data: DataDict) -> t.Dict[str, str]:
    """
    Return dictionary with ASCII-clean keys and values.
    """

    outdata = OrderedDict()
    for key, value in data.items():
        key = ascii_clean(key).strip()
        value = ascii_clean(value).strip()
        outdata[key] = value
    return outdata


def encode_ntfy_fields(data: DataDict) -> t.Dict[str, str]:
    """
    Return dictionary suitable for submitting to the ntfy HTTP API using HTTP headers.

    - The field values for `title`, `message` and `tags` are encoded using RFC 2047, aka.
      MIME Message Header Extensions for Non-ASCII Text.

    - The other field values will be stripped from any special characters to be ASCII-clean.

    Appendix

    When using RFC 2047, two encodings are possible. The Python implementation cited below
    seems to use the "Q" encoding scheme by default.

    4.1 The "B" encoding is identical to the "BASE64" encoding defined by RFC 2045.
    4.2 The "Q" encoding is similar to the "Quoted-Printable" content-transfer-encoding
        defined in RFC 2045.  It is designed to allow text containing mostly ASCII
        characters to be decipherable on an ASCII terminal without decoding.

    The Python email package supports the standards RFC 2045, RFC 2046, RFC 2047, and
    RFC 2231 in its `email.header` and `email.charset` modules.

    - https://datatracker.ietf.org/doc/html/rfc2047#section-2
    - https://datatracker.ietf.org/doc/html/rfc2047#section-4
    - https://docs.python.org/3/library/email.header.html
    """

    outdata = OrderedDict()
    for key, value in data.items():
        key = ascii_clean(key).strip()
        if key in NTFY_RFC2047_FIELDS:
            value = encode_rfc2047(value)
        else:
            value = ascii_clean(value)
        outdata[key] = value
    return outdata


def dict_with_titles(data: t.Dict[str, str]) -> t.Dict[str, str]:
    """
    Return dictionary with each key title-cased, i.e. uppercasing the first letter.

    >>> {"foo": "bar"}
    {"Foo": "bar"}
    """
    outdata = OrderedDict()
    for key, value in data.items():
        outdata[key.title()] = value
    return outdata



# -*- coding: utf-8 -*-
# hiveeyes-schwarmalarm function extensions for mqttwarn
# https://hiveeyes.org/docs/system/schwarmalarm-mqttwarn.html
from builtins import object
from builtins import str
import re
import json
from datetime import datetime
from pprint import pformat
from collections import deque, defaultdict


# ------------------------------------------
#                  Synopsis
# ------------------------------------------
# Setup dependencies::
#
#   pip install xmpppy==0.5.0rc1 jinja2==2.8
#
# Run mqttwarn::
#
#   export MQTTWARNINI=examples/hiveeyes/hiveeyes.ini
#   mqttwarn
#
# Trigger an alarm by simulating a weight loss event::
#
#   echo '{"wght2": 43.0}' | mosquitto_pub -t hiveeyes/demo/area-42/beehive-1/data.json -l
#   echo '{"wght2": 42.0}' | mosquitto_pub -t hiveeyes/demo/area-42/beehive-1/data.json -l


# ------------------------------------------
#                Configuration
# ------------------------------------------

# Ruleset for monitoring measurement values is defined
# by mapping field names to delta threshold values.
monitoring_rules = {

    # Let's trigger an alarm on a weight loss
    # of 750g or more between two measurements
    'wght2':  {'threshold': 0.75, 'unit': 'kilo'},

    # For testing the machinery by
    # monitoring a clock signal
    'second': {'threshold': 0.5, 'unit': 'second'},
    }

# Trigger a data-loss event when not receiving any fresh
# data after 20 minutes from a device already seen before.
data_loss_timeout = 20 * 60

# Just a hack to give the anonymous data source of "Labhive One" a better name
dashboard_aliases = {
    '25a0e5df-9517-405b-ab14-cb5b514ac9e8': 'hiveeyes-labs-wedding',
}


# ------------------------------------------
#                   Main
# ------------------------------------------

# Data structure for remembering measurement values across a variable window size.
class HistoricData(object):

    # How many data points to remember
    backlog = 20

    def __init__(self):
        series_factory = lambda: deque(maxlen=HistoricData.backlog)

        # Some data containers for remembering ...

        # Inbound data payload fragments
        self.fragments = defaultdict(series_factory)

        # All attributes aggregated across consecutive data payload fragments
        # This is actually used for threshold comparison
        self.history = defaultdict(series_factory)

        # Arbitrary node states like "data-loss"
        self.states = defaultdict(dict)

        # Most recent moment of data acquisition per node
        self.moments = {}

# A single data container instance to rule them all
hdata = HistoricData()


def format_passthrough(data):
    """
    Stringify complete transformation data from mqttwarn
    to assist debugging as a pass-through formatter.
    """
    return str(data)


def hiveeyes_topic_to_topology(topic):
    """
    Split Hiveeyes MQTT topic path into segments for
    enriching transformation data inside mqttwarn.
    """
    if type(topic) == str:
        try:
            pattern = r'^(?P<realm>.+?)/(?P<network>.+?)/(?P<gateway>.+?)/(?P<node>.+?)/(?P<field>.+?)$'
            p = re.compile(pattern)
            m = p.match(topic)
            topology = m.groupdict()
        except:
            topology = {}
        return topology
    return None


def hiveeyes_more_data(topic, data, srv):
    """
    Add more data to transformation data object, used later when formatting the outgoing message.
    """

    if not (topic.endswith('data.json') or topic.endswith('message-json')):
        return

    message = data['payload']

    # Decode message (redundant with "hiveeyes_schwarmalarm_filter")
    mdata = dict(json.loads(message).items())
    tdata = mdata.copy()
    tdata.update(hiveeyes_topic_to_topology(topic))

    # Compute bookkeeping key for multi-tenancy
    origin = '{realm}/{network}/{gateway}/{node}'.format(**tdata)

    condition = {'current': None, 'previous': None}

    # Get current and previous measurements
    fragments = condition.copy()
    try:
        fragments['current'] = hdata.fragments[origin][-1]
    except IndexError:
        pass
    try:
        fragments['previous'] = hdata.fragments[origin][-2]
    except IndexError:
        pass

    history = condition.copy()
    try:
        history['current'] = hdata.history[origin][-1]
    except IndexError:
        pass
    try:
        history['previous'] = hdata.history[origin][-2]
    except IndexError:
        pass

    # Compute Grafana dashboard name from simple map, fall back
    # to network identifier as this is the default of Kotori's GrafanaManager.
    dashboard = dashboard_aliases.get(tdata['network'], tdata['network'])

    more_data = {

        # current/previous data
        'fragments': fragments,
        'history': history,

        # more enrichment
        'dashboard': dashboard,

        # utilities
        'pformat': pformat,
    }

    # Also push the "event" state into the template machinery
    # This is used for displaying a preformatted description message to humans
    if 'event' in hdata.states[origin]:
        more_data.update(hdata.states[origin]['event'])

    return more_data


def hiveeyes_schwarmalarm_filter(topic, message):
    """
    Custom filter function to compare two consecutive values
    to trigger notification only if delta is greater threshold.
    Also, perform data-loss bookkeeping.
    """

    def cmp(a, b):
        """
        https://docs.python.org/3.0/whatsnew/3.0.html#ordering-comparisons
        """
        return (a > b) - (a < b)

    if not (topic.endswith('data.json') or topic.endswith('message-json')):
        return True

    # Decode message (redundant with "hiveeyes_more_data")
    mdata = dict(json.loads(message).items())
    tdata = mdata.copy()
    tdata.update(hiveeyes_topic_to_topology(topic))

    # Compute bookkeeping key for multi-tenancy
    origin = '{realm}/{network}/{gateway}/{node}'.format(**tdata)

    # Data-loss bookkeeping
    hdata.moments[origin] = datetime.utcnow()

    # Get most recent measurement from series
    try:
        history = hdata.history[origin][-1].copy()
    except IndexError:
        history = {}

    alarm = False
    for field in monitoring_rules.keys():

        # Skip if monitored field is not in data payload
        if field not in tdata:
            continue

        # Read current value and appropriate threshold
        current = tdata[field]
        threshold = monitoring_rules[field]['threshold']

        # Compare current with former value and set
        # semaphore if delta is greater threshold
        if field in history:
            previous = history[field]
            delta = float(current) - float(previous)
            delta_nosign = abs(delta)
            if delta_nosign >= threshold:

                # Set event state outcome / semaphore
                alarm = True

                # Format description message for displaying to humans, e.g.
                # »Sensor value "wght2" gained 1.42 kilos.«
                unit = monitoring_rules[field].get('unit', 'point')
                verb = delta != 0 and (cmp(delta, 0) == 1 and 'gained' or 'lost') or 'did not change'
                description = 'Sensor value "{field}" {verb} {delta_nosign} {unit}s.'.format(
                    field=field, verb=verb, delta_nosign=delta_nosign, unit=unit)
                hdata.states[origin]['event'] = {'field': field, 'delta': delta, 'description': description}
            else:
                # Reset event state
                if 'event' in hdata.states[origin]:
                    del hdata.states[origin]['event']

    # Remember current values for next round
    hdata.fragments[origin].append(mdata)

    history.update(mdata)
    hdata.history[origin].append(history)

    # The filter function should return True if the message should
    # be suppressed, or False if the message should be processed.
    #return False
    return not alarm


def hiveeyes_dataloss_monitor(srv):
    """
    Periodic thread monitoring all seen devices for data loss events.
    Will submit appropriate data loss notifications using mqttwarn services.
    """

    # Get hold of our own PeriodicThread object to read "period" value
    periodic_threads = srv.mwcore['ptlist']
    me = periodic_threads['hiveeyes_dataloss_monitor']
    interval = me.period

    # Inform operator about this activity
    srv.logging.info('Checking for data-loss (each {interval} seconds)'.format(**locals()))

    # Get hold of mqttwarn method to dispatch notification messages on data loss events
    send_to_targets = srv.mwcore['send_to_targets']

    now = datetime.utcnow()

    # List of all seen data origins (devices).
    # The identifiers are made of essential parts of the MQTT topic.
    origins = hdata.moments.keys()

    # Iterate all seen devices
    for origin in origins:

        # Signal no data loss as a default
        hdata.states[origin].setdefault('data-loss', False)

        # Get timestamp of last valid measurement
        last_moment = hdata.moments[origin]

        # Compute duration since last valid measurement
        delta = now - last_moment

        # Check if duration is longer than user-defined threshold
        if delta.total_seconds() >= data_loss_timeout:

            # Only act if not already being in state of data loss
            if 'data-loss' not in hdata.states[origin] or not hdata.states[origin]['data-loss']:

                # Set data loss state
                hdata.states[origin]['data-loss'] = True

                # Send out data loss notification
                data = {'description': 'Detected data loss.'}
                send_to_targets(
                    section = 'hiveeyes-schwarmalarm',
                    topic   = '{origin}/notification.json'.format(origin=origin),
                    payload = json.dumps(data))

        else:
            # Reset data loss state
            hdata.states[origin]['data-loss'] = False

    # TODO: Republish data loss notification to MQTT bus
    #srv.mqttc.publish('data-loss-topic'.format(**locals()), 'data-loss')


# ------------------------------------------
#                   Setup
# ------------------------------------------
# Duplicate of mqttwarn helper method for loading service plugins
# http://code.davidjanes.com/blog/2008/11/27/how-to-dynamically-load-python-code/
def load_module(path):
    import imp
    import hashlib
    try:
        fp = open(path, 'rb')
        return imp.load_source(hashlib.md5(path).hexdigest(), path, fp)
    finally:
        try:
            fp.close()
        except:
            pass

# Mitigate "AttributeError: '_ssl._SSLSocket' object has no attribute 'issuer'"
service_xmpp = load_module('services/xmpp.py')
service_xmpp.xmpppy_monkeypatch_ssl()

